"""Implements the base rule class.

Rules crawl through the trees returned by the parser and evaluate particular
rules.

The intent is that it should be possible for the rules to be expressed
as simply as possible, with as much of the complexity abstracted away.

The evaluation function should take enough arguments that it can evaluate
the position of the given segment in relation to its neighbors, and that
the segment which finally "triggers" the error, should be the one that would
be corrected OR if the rule relates to something that is missing, then it
should flag on the segment FOLLOWING, the place that the desired element is
missing.
"""

import bdb
import copy
import fnmatch
import logging
import pathlib
import re
import threading
from collections import defaultdict, namedtuple
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    DefaultDict,
    Optional,
    Union,
)

import regex

from sqlfluff.core.errors import SQLFluffUserError, SQLLintError
from sqlfluff.core.helpers.string import split_comma_separated_string
from sqlfluff.core.parser import BaseSegment, RawSegment
from sqlfluff.core.plugin.host import is_main_process, plugins_loaded
from sqlfluff.core.rules.config_info import ConfigInfo, get_config_info
from sqlfluff.core.rules.context import RuleContext
from sqlfluff.core.rules.crawlers import BaseCrawler
from sqlfluff.core.rules.fix import LintFix
from sqlfluff.core.templaters.base import TemplatedFile

# Best solution for generic types on older python versions
# https://github.com/python/typeshed/issues/7855
if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.config import FluffConfig
    from sqlfluff.core.dialects import Dialect
    from sqlfluff.core.plugin.hookspecs import PluginSpec
    from sqlfluff.core.rules.noqa import IgnoreMask

    _LoggerAdapter = logging.LoggerAdapter[logging.Logger]
else:
    _LoggerAdapter = logging.LoggerAdapter

# The ghost of a rule (mostly used for testing)
RuleGhost = namedtuple("RuleGhost", ["code", "name", "description"])

# Instantiate the rules logger
rules_logger = logging.getLogger("sqlfluff.rules")

linter_logger: logging.Logger = logging.getLogger("sqlfluff.linter")


class RuleLoggingAdapter(_LoggerAdapter):
    """A LoggingAdapter for rules which adds the code of the rule to it."""

    def process(self, msg: str, kwargs: Any) -> tuple[str, Any]:
        """Add the code element to the logging message before emit."""
        return "[{}] {}".format(self.extra["code"] if self.extra else "", msg), kwargs


class LintResult:
    """A class to hold the results of a rule evaluation.

    Args:
        anchor (:obj:`BaseSegment`, optional): A segment which represents
            the *position* of the problem. NB: Each fix will also hold
            its own reference to position, so this position is mostly for
            alerting the user to where the *problem* is.
        fixes (:obj:`list` of :obj:`LintFix`, optional): An array of any
            fixes which would correct this issue. If not present then it's
            assumed that this issue will have to manually fixed.
        memory (:obj:`dict`, optional): An object which stores any working
            memory for the rule. The `memory` returned in any `LintResult`
            will be passed as an input to the next segment to be crawled.
        description (:obj:`str`, optional): A description of the problem
            identified as part of this result. This will override the
            description of the rule as what gets reported to the user
            with the problem if provided.
        source (:obj:`str`, optional): A string identifier for what
            generated the result. Within larger libraries like reflow this
            can be useful for tracking where a result came from.

    """

    def __init__(
        self,
        anchor: Optional[BaseSegment] = None,
        fixes: Optional[list["LintFix"]] = None,
        memory: Optional[Any] = None,
        description: Optional[str] = None,
        source: Optional[str] = None,
    ):
        # An anchor of none, means no issue
        self.anchor = anchor
        # Fixes might be blank
        self.fixes = fixes or []
        # Memory is passed back in the linting result
        self.memory = memory
        # store a description_override for later
        self.description = description
        # Optional code for where the result came from
        self.source: str = source or ""

    def __repr__(self) -> str:
        if not self.anchor:
            return "LintResult(<empty>)"
        # The "F" at the end is short for "fixes", to indicate how many there are.
        fix_coda = f"+{len(self.fixes)}F" if self.fixes else ""
        if self.description:
            if self.source:
                return (
                    f"LintResult({self.description} [{self.source}]"
                    f": {self.anchor}{fix_coda})"
                )
            return f"LintResult({self.description}: {self.anchor}{fix_coda})"
        return f"LintResult({self.anchor}{fix_coda})"

    def to_linting_error(self, rule: "BaseRule") -> Optional[SQLLintError]:
        """Convert a linting result to a :exc:`SQLLintError` if appropriate."""
        if self.anchor:
            # Allow description override from the LintResult
            description = self.description or rule.description
            return SQLLintError(
                rule=rule,
                segment=self.anchor,
                fixes=self.fixes,
                description=description,
            )

        return None


EvalResultType = Union[LintResult, list[LintResult], None]


class RuleMetaclass(type):
    """The metaclass for rules.

    This metaclass provides provides auto-enrichment of the
    rule docstring so that examples, groups, aliases and
    names are added.

    The reason we enrich the docstring is so that it can be
    picked up by autodoc and all be displayed in the sqlfluff
    docs.
    """

    # Precompile the regular expressions
    _doc_search_regex: ClassVar = re.compile(
        "(\\s{4}\\*\\*Anti-pattern\\*\\*|\\s{4}\\.\\. note::|"
        "\\s\\s{4}\\*\\*Configuration\\*\\*)",
        flags=re.MULTILINE,
    )
    _valid_classname_regex: ClassVar = regex.compile(
        r"Rule_?([A-Z]{1}[a-zA-Z]+)?_([A-Z0-9]{4})"
    )
    _valid_rule_name_regex: ClassVar = regex.compile(r"[a-z][a-z\.\_]+")

    @staticmethod
    def _populate_code_and_description(
        name: str, class_dict: dict[str, Any]
    ) -> dict[str, Any]:
        """Extract and validate the rule code & description.

        We expect that rules are defined as classes with the name `Rule_XXXX`
        where `XXXX` is of the form `LLNN`, where L is a letter and N is a
        two digit number. For backward compatibility we also still support
        the legacy format of LNNN i.e. a single letter and three digit number.

        The two letters should be indicative of the grouping and focus of
        the rule. e.g. capitalisation rules have the code CP for CaPitalisation.

        If this receives classes by any other name, then it will raise a
        :exc:`ValueError`.
        """
        rule_name_match = RuleMetaclass._valid_classname_regex.match(name)
        # Validate the name
        if not rule_name_match:  # pragma: no cover
            raise SQLFluffUserError(
                f"Tried to define rule class with "
                f"unexpected format: {name}. Format should be: "
                "'Rule_PluginName_LL23' (for plugins) or "
                "`Rule_LL23` (for core rules)."
            )

        plugin_name, code = rule_name_match.groups()
        # If the docstring is multiline, then we extract just summary.
        description = class_dict["__doc__"].replace("``", "'").split("\n")[0]
        if plugin_name:
            code = f"{plugin_name}_{code}"

        class_dict["code"] = code
        class_dict["description"] = description

        return class_dict

    @staticmethod
    def _populate_docstring(name: str, class_dict: dict[str, Any]) -> dict[str, Any]:
        """Enrich the docstring in the class_dict.

        This takes the various defined values in the BaseRule class
        and uses them to populate documentation in the final class
        docstring so that it can be displayed in the sphinx docs.
        """
        # Ensure that there _is_ a docstring.
        assert "__doc__" in class_dict, (
            f"Tried to define rule {name!r} without docstring."
        )

        # Build up a buffer of entries to add to the docstring.
        fix_docs = (
            "    This rule is ``sqlfluff fix`` compatible.\n\n"
            if class_dict.get("is_fix_compatible", False)
            else ""
        )
        name_docs = (
            f"    **Name**: ``{class_dict['name']}``\n\n"
            if class_dict.get("name", "")
            else ""
        )
        alias_docs = (
            ("    **Aliases**: ``" + "``, ``".join(class_dict["aliases"]) + "``\n\n")
            if class_dict.get("aliases", [])
            else ""
        )
        groups_docs = (
            ("    **Groups**: ``" + "``, ``".join(class_dict["groups"]) + "``\n\n")
            if class_dict.get("groups", [])
            else ""
        )

        config_docs = ""

        # NOTE: We should only validate and add config keywords
        # into the docstring if the plugin loading methods have
        # fully completed (i.e. plugins_loaded.get() is True).
        if (
            name == "BaseRule"
            or not is_main_process.get()
            or threading.current_thread() is not threading.main_thread()
        ):
            # Except if it's the base rule, or we're not in the main process/thread
            # in which case we shouldn't try and alter the docstrings anyway.
            # NOTE: The order of imports within child threads/processes is less
            # controllable, and so we should just avoid checking whether plugins
            # are already loaded.
            # NOTE: We also suppress this warning in non-main threads (e.g. the
            # multiprocessing pool's _handle_results thread in the main process)
            # because Python 3.11 threads start with ContextVar defaults rather
            # than a copy of the spawning thread's context, so plugins_loaded is
            # False even though plugins are fully loaded in the main thread.
            pass
        elif not plugins_loaded.get():
            # Show a warning if a plugin has their imports set up in a suboptimal
            # way. The example plugin imports the rules in both ways, to test the
            # triggering of this warning.
            rules_logger.warning(
                f"Rule {name!r} has been imported before all plugins "
                "have been fully loaded. For best performance, plugins "
                "should import any rule definitions within their `get_rules()` "
                "method. Please update your plugin to remove this warning. See: "
                "https://docs.sqlfluff.com/en/stable/perma/plugin_dev.html"
            )
        elif class_dict.get("config_keywords", []):
            config_docs = "\n    **Configuration**\n"
            config_info = get_config_info()
            for keyword in sorted(class_dict["config_keywords"]):
                try:
                    info_dict = config_info[keyword]
                except KeyError:  # pragma: no cover
                    # NOTE: For rule developers, please define config info values
                    # within the specific rule bundle rather than in the central
                    # `config_info` package unless the value is necessary for
                    # multiple rules.
                    raise KeyError(
                        "Config value {!r} for rule {} is not configured in "
                        "`config_info`.".format(keyword, name)
                    )
                config_docs += "\n    * ``{}``: {}".format(
                    keyword, info_dict["definition"]
                )
                if (
                    config_docs[-1] != "."
                    and config_docs[-1] != "?"
                    and config_docs[-1] != "\n"
                ):
                    config_docs += "."
                if "validation" in info_dict:
                    config_docs += " Must be one of ``{}``.".format(
                        info_dict["validation"]
                    )
            config_docs += "\n"

        all_docs = fix_docs + name_docs + alias_docs + groups_docs + config_docs
        # Modify the docstring using the search regex.
        class_dict["__doc__"] = RuleMetaclass._doc_search_regex.sub(
            f"\n\n{all_docs}\n\n\\1", class_dict["__doc__"], count=1
        )
        # If the inserted string is not now in the docstring - append it on
        # the end. This just means the regex didn't find a better place to
        # put it.
        if all_docs not in class_dict["__doc__"]:
            class_dict["__doc__"] += f"\n\n{all_docs}"

        # Return the modified class_dict
        return class_dict

    def __new__(
        mcs,
        name: str,
        bases: list["BaseRule"],
        class_dict: dict[str, Any],
    ) -> "RuleMetaclass":
        """Generate a new class."""
        # Optionally, groups may be inherited. At this stage of initialisation
        # they won't have been. Check parent classes if they exist.
        # names, aliases and description are less appropriate to inherit.
        # NOTE: This applies in particular to CP02, which inherits all groups
        # from CP01. If we don't do this, those groups don't show in the docs.
        for base in reversed(bases):
            if "groups" in class_dict:
                break
            elif base.groups:
                class_dict["groups"] = base.groups
                break

        # If the rule doesn't itself define `config_keywords`, check the parent
        # classes for them. If we don't do this then they'll still be available to
        # the rule, but they won't appear in the docs.
        for base in reversed(bases):
            if "config_keywords" in class_dict:
                break
            elif base.config_keywords:
                class_dict["config_keywords"] = base.config_keywords
                break

        class_dict = RuleMetaclass._populate_docstring(name, class_dict)
        # Don't try and infer code and description for the base classes
        if name not in ("BaseRule",):
            class_dict = RuleMetaclass._populate_code_and_description(name, class_dict)
        # Validate rule names
        rule_name = class_dict.get("name", "")
        if rule_name:
            if not RuleMetaclass._valid_rule_name_regex.match(rule_name):
                raise SQLFluffUserError(
                    f"Tried to define rule with unexpected "
                    f"name format: {rule_name}. Rule names should be lowercase "
                    "and snake_case with optional `.` characters to indicate "
                    "a namespace or grouping. e.g. `layout.spacing`."
                )

        # Use the stock __new__ method now we've adjusted the docstring.
        # There are no overload variants of type.__new__ that are compatible, so
        # we ignore type checking in this case.
        return super().__new__(mcs, name, bases, class_dict)  # type: ignore


class BaseRule(metaclass=RuleMetaclass):
    """The base class for a rule.

    Args:
        code (:obj:`str`): The identifier for this rule, used in inclusion
            or exclusion.
        description (:obj:`str`): A human readable description of what this
            rule does. It will be displayed when any violations are found.

    """

    _check_docstring = True
    _works_on_unparsable = True
    _adjust_anchors = False
    targets_templated = False
    # Some fix routines do their own checking for whether their fixes
    # are safe around templated elements. For those - the default
    # safety checks might be inappropriate. In those cases, set
    # template_safe_fixes to True.
    template_safe_fixes = False

    # Config settings supported for this rule.
    # See config_info.py for supported values.
    config_keywords: list[str] = []
    # Lint loop / crawl behavior. When appropriate, rules can (and should)
    # override these values to make linting faster.
    crawl_behaviour: BaseCrawler
    # Rules can override this to specify "post". "Post" rules are those that are
    # not expected to trigger any downstream rules, e.g. capitalization fixes.
    # They run on two occasions:
    # - On the first pass of the main phase
    # - In a second linter pass after the main phase
    lint_phase = "main"
    # Groups attribute to be overwritten.
    groups: tuple[str, ...] = ()
    # Name attribute to be overwritten.
    # NOTE: for backward compatibility we should handle the case
    # where no name is set gracefully.
    name: str = ""
    # Optional set of aliases for the rule. Most often used for old codes which
    # referred to this rule.
    aliases: tuple[str, ...] = ()

    # NOTE: code and description are provided here as hints, but should not
    # be set directly. They are set automatically by the metaclass based on
    # the class _name_ when defined.
    code: str
    description: str

    # Should we document this rule as fixable? Used by the metaclass to add
    # a line to the docstring.
    is_fix_compatible = False

    # Add comma separated string to Base Rule to ensure that it uses the same
    # Configuration that is defined in the Config.py file
    split_comma_separated_string = staticmethod(split_comma_separated_string)

    def __init__(self, code: str, description: str, **kwargs: Any) -> None:
        self.description = description
        self.code = code
        # kwargs represents the config passed to the rule. Add all kwargs as class
        # attributes so they can be accessed in rules which inherit from this class
        for key, value in kwargs.items():
            self.__dict__[key] = value

        # We also define a custom logger here, which also includes the code
        # of the rule in the logging.
        self.logger = RuleLoggingAdapter(rules_logger, {"code": code})
        # Validate that declared configuration options exist
        for keyword in self.config_keywords:
            if keyword not in kwargs.keys():
                raise ValueError(
                    (
                        "Unrecognized config '{}' for Rule {}. If this "
                        "is a new option, please add it to "
                        "`default_config.cfg` or plugin specific config."
                    ).format(keyword, code)
                )

    @classmethod
    def get_config_ref(cls) -> str:
        """Return the config lookup ref for this rule.

        If a `name` is defined, it's the name - otherwise the code.

        The name is a much more understandable reference and so makes config
        files more readable. For backward compatibility however we also support
        the rule code for those without names.
        """
        return cls.name if cls.name else cls.code

    def _eval(self, context: RuleContext) -> EvalResultType:
        """Evaluate this rule against the current context.

        This should indicate whether a linting violation has occurred and/or
        whether there is something to remember from this evaluation.

        Note that an evaluate function should always accept `**kwargs`, but
        if it relies on any available kwargs, it should explicitly call
        them out at definition.

        Returns:
            :obj:`LintResult`, list of :obj:`LintResult` or :obj:`None`.

        The reason that this method is called :meth:`_eval` and not `eval` is
        a bit of a hack with sphinx autodoc, to make it so that the rule
        documentation auto-generates nicely.

        """
        raise NotImplementedError(
            (
                "{} has not had its `eval` function defined. This is a problem "
                "with the rule setup."
            ).format(self.__class__.__name__)
        )  # pragma: no cover

    @staticmethod
    def _rust_rules_enabled(config: "FluffConfig") -> bool:
        """Whether Rust-native rule detection is enabled (core.use_rust_rules)."""
        return config.get_section(["core", "use_rust_rules"]) in (
            "auto",
            "Auto",
            "AUTO",
            True,
            "True",
            "true",
            1,
        )

    def _eval_rust(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Optionally evaluate the whole rule natively in Rust.

        Override in a rule to compute *all* of its ``LintResult``s in one pass
        over the Rust arena (``context.segment._rs_tree`` — ``context`` is the
        root context here), instead of the per-segment Python ``_eval`` crawl.

        Return ``None`` (the default) to use the Python path. Only consulted
        when ``core.use_rust_rules`` is enabled and the parse produced an arena;
        the returned results are post-processed identically to ``_eval``
        results, so a rule should also return ``None`` for any configuration it
        does not implement natively (so it falls back to Python).
        """
        return None

    def crawl(
        self,
        tree: BaseSegment,
        dialect: "Dialect",
        fix: bool,
        templated_file: Optional["TemplatedFile"],
        ignore_mask: Optional["IgnoreMask"],
        fname: Optional[str],
        config: "FluffConfig",
    ) -> tuple[
        list[SQLLintError],
        tuple[RawSegment, ...],
        list[LintFix],
        Optional[dict[str, Any]],
    ]:
        """Run the rule on a given tree.

        Returns:
            A tuple of (vs, raw_stack, fixes, memory)

        """
        root_context = RuleContext(
            dialect=dialect,
            fix=fix,
            templated_file=templated_file,
            path=pathlib.Path(fname) if fname else None,
            segment=tree,
            config=config,
        )
        vs: list[SQLLintError] = []
        fixes: list[LintFix] = []

        # Experimental: whole-rule Rust-native dispatch (opt-in via
        # core.use_rust_rules). When enabled and the parse produced an arena, a
        # rule may compute all of its results in one pass over root._rs_tree
        # (see _eval_rust) instead of the per-segment Python crawl. The results
        # are post-processed identically to _eval results. Rules without a Rust
        # path — or parses with no arena (Python parser) — fall through to the
        # Python crawl below.
        if self._rust_rules_enabled(config) and getattr(tree, "_rs_tree", None):
            try:
                rust_results = self._eval_rust(root_context)
            except (bdb.BdbQuit, KeyboardInterrupt):  # pragma: no cover
                raise
            except Exception as e:
                # Mirror the per-segment handler below: a Rust binding/mapping
                # error must not abort linting the whole file — surface it as a
                # recoverable SQLLintError and stop this rule.
                self.logger.critical(
                    (
                        f"Applying rule {self.code} (rust) to {fname!r} "
                        f"threw an Exception: {e}"
                        if fname
                        else f"Applying rule {self.code} (rust) threw an Exception: {e}"
                    ),
                    exc_info=True,
                )
                assert root_context.segment.pos_marker
                exception_line, _ = root_context.segment.pos_marker.source_position()
                self._log_critical_errors(e)
                vs.append(
                    SQLLintError(
                        rule=self,
                        segment=root_context.segment,
                        fixes=[],
                        description=(
                            f"Unexpected exception: {str(e)};\n"
                            "Could you open an issue at "
                            "https://github.com/sqlfluff/sqlfluff/issues ?\n"
                            "You can ignore this exception for now, by adding "
                            f"'-- noqa: {self.code}' at the end\n"
                            f"of line {exception_line}\n"
                        ),
                    )
                )
                return vs, root_context.raw_stack, fixes, root_context.memory
            if rust_results is not None:
                # Distinct names from the per-segment loop below so each keeps
                # its own inferred type (mypy treats reused names as redefs).
                for rust_res in rust_results:
                    rust_lerrs: list[SQLLintError] = []
                    rust_fixes: list[LintFix] = []
                    self._adjust_anchors_for_fixes(root_context, rust_res)
                    self._process_lint_result(
                        rust_res,
                        templated_file,
                        ignore_mask,
                        rust_lerrs,
                        rust_fixes,
                        tree,
                    )
                    vs += rust_lerrs
                    fixes += rust_fixes
                return vs, root_context.raw_stack, fixes, root_context.memory

        # Propagates memory from one rule _eval() to the next.
        memory = root_context.memory
        context = root_context
        for context in self.crawl_behaviour.crawl(root_context):
            try:
                context.memory = memory
                res = self._eval(context=context)
            except (bdb.BdbQuit, KeyboardInterrupt):  # pragma: no cover
                raise
            # Any exception at this point would halt the linter and
            # cause the user to get no results
            except Exception as e:
                # If a filename is present, include it in the critical exception.
                self.logger.critical(
                    (
                        f"Applying rule {self.code} to {fname!r} "
                        f"threw an Exception: {e}"
                        if fname
                        else f"Applying rule {self.code} threw an Exception: {e}"
                    ),
                    exc_info=True,
                )
                assert context.segment.pos_marker
                exception_line, _ = context.segment.pos_marker.source_position()
                self._log_critical_errors(e)
                vs.append(
                    SQLLintError(
                        rule=self,
                        segment=context.segment,
                        fixes=[],
                        description=(
                            f"Unexpected exception: {str(e)};\n"
                            "Could you open an issue at "
                            "https://github.com/sqlfluff/sqlfluff/issues ?\n"
                            "You can ignore this exception for now, by adding "
                            f"'-- noqa: {self.code}' at the end\n"
                            f"of line {exception_line}\n"
                        ),
                    )
                )
                return vs, context.raw_stack, fixes, context.memory

            new_lerrs: list[SQLLintError] = []
            new_fixes: list[LintFix] = []

            if res is None or res == []:
                # Assume this means no problems (also means no memory)
                pass
            elif isinstance(res, LintResult):
                # Extract any memory
                memory = res.memory
                self._adjust_anchors_for_fixes(context, res)
                self._process_lint_result(
                    res, templated_file, ignore_mask, new_lerrs, new_fixes, tree
                )
            elif isinstance(res, list) and all(
                isinstance(elem, LintResult) for elem in res
            ):
                # Extract any memory from the *last* one, assuming
                # it was the last to be added
                memory = res[-1].memory
                for elem in res:
                    self._adjust_anchors_for_fixes(context, elem)
                    self._process_lint_result(
                        elem, templated_file, ignore_mask, new_lerrs, new_fixes, tree
                    )
            else:  # pragma: no cover
                raise TypeError(
                    "Got unexpected result [{!r}] back from linting rule: {!r}".format(
                        res, self.code
                    )
                )

            for lerr in new_lerrs:
                self.logger.info("!! Violation Found: %r", lerr.description)
            if new_fixes:
                if not self.is_fix_compatible:  # pragma: no cover
                    rules_logger.error(
                        f"Rule {self.code} returned a fix but is not documented as "
                        "`is_fix_compatible`, you may encounter unusual fixing "
                        "behaviour. Report this a bug to the developer of this rule."
                    )
                for lfix in new_fixes:
                    self.logger.info("!! Fix Proposed: %r", lfix)

            # Consume the new results
            vs += new_lerrs
            fixes += new_fixes
        return vs, context.raw_stack if context else tuple(), fixes, context.memory

    # HELPER METHODS --------
    @staticmethod
    def _log_critical_errors(error: Exception) -> None:  # pragma: no cover
        """This method is monkey patched into a "raise" for certain tests."""
        pass

    def _process_lint_result(
        self,
        res: LintResult,
        templated_file: Optional[TemplatedFile],
        ignore_mask: Optional["IgnoreMask"],
        new_lerrs: list[SQLLintError],
        new_fixes: list[LintFix],
        root: BaseSegment,
    ) -> None:
        # Unless the rule declares that it's already template safe. Do safety
        # checks.
        if not self.template_safe_fixes:
            self.discard_unsafe_fixes(res, templated_file)
        lerr = res.to_linting_error(rule=self)
        if not lerr:
            return None
        if ignore_mask:
            if not ignore_mask.ignore_masked_violations([lerr]):
                return None

        # Check whether this should be filtered out for being unparsable.
        # To do that we check the parents of the anchors (of the violation
        # and fixes) against the filter in the crawler.
        # NOTE: We use `.passes_filter` here to do the test for unparsable
        # to avoid duplicating code because that test is already implemented
        # there.
        anchors = [lerr.segment] + [fix.anchor for fix in lerr.fixes]
        for anchor in anchors:
            if not self.crawl_behaviour.passes_filter(anchor):  # pragma: no cover
                # NOTE: This clause is untested, because it's a hard to produce
                # edge case. The latter clause is much more likely.
                linter_logger.info(
                    "Fix skipped due to anchor not passing filter: %s", anchor
                )
                return None

            parent_stack = root.path_to(anchor)
            if not all(
                self.crawl_behaviour.passes_filter(ps.segment) for ps in parent_stack
            ):
                linter_logger.info(
                    "Fix skipped due to parent of anchor not passing filter: %s",
                    [ps.segment for ps in parent_stack],
                )
                return None

        new_lerrs.append(lerr)
        new_fixes.extend(res.fixes)

    @staticmethod
    def filter_meta(
        segments: Sequence[BaseSegment], keep_meta: bool = False
    ) -> tuple[BaseSegment, ...]:
        """Filter the segments to non-meta.

        Or optionally the opposite if keep_meta is True.
        """
        buff = []
        for elem in segments:
            if elem.is_meta is keep_meta:
                buff.append(elem)
        return tuple(buff)

    @classmethod
    def get_parent_of(
        cls, segment: BaseSegment, root_segment: BaseSegment
    ) -> Optional[BaseSegment]:  # pragma: no cover TODO?
        """Return the segment immediately containing segment.

        NB: This is recursive.

        Args:
            segment: The segment to look for.
            root_segment: Some known parent of the segment
                we're looking for (although likely not the
                direct parent in question).

        """
        if segment in root_segment.segments:
            return root_segment
        elif root_segment.segments:
            # try each of the subsegments
            for sub in root_segment.segments:
                p = cls.get_parent_of(segment, sub)
                if p:
                    return p
        # Not directly in the segment and
        # no subsegments to check. Return None.
        return None

    @staticmethod
    def discard_unsafe_fixes(
        lint_result: LintResult, templated_file: Optional[TemplatedFile]
    ) -> None:
        """Remove (discard) LintResult fixes if they are "unsafe".

        By removing its fixes, a LintResult will still be reported, but it
        will be treated as _unfixable_.
        """
        if not lint_result.fixes or not templated_file:
            return

        # Check for fixes that touch templated code.
        for fix in lint_result.fixes:
            if fix.has_template_conflicts(templated_file):
                linter_logger.info(
                    "      * Discarding fixes that touch templated code: %s",
                    lint_result.fixes,
                )
                lint_result.fixes = []
                return

        # Issue 3079: Fixes that span multiple template blocks are bad. Don't
        # permit them.
        block_indices: set[int] = set()
        for fix in lint_result.fixes:
            fix_slices = fix.get_fix_slices(templated_file, within_only=True)
            for fix_slice in fix_slices:
                # Ignore fix slices that exist only in the source. For purposes
                # of this check, it's not meaningful to say that a fix "touched"
                # one of these.
                if not fix_slice.is_source_only_slice():
                    block_indices.add(fix_slice.block_idx)
        if len(block_indices) > 1:
            linter_logger.info(
                "      * Discarding fixes that span multiple template blocks: %s",
                lint_result.fixes,
            )
            lint_result.fixes = []
            return

    @classmethod
    def _adjust_anchors_for_fixes(
        cls, context: RuleContext, lint_result: LintResult
    ) -> None:
        """Makes simple fixes to the anchor position for fixes.

        Some rules return fixes where the anchor is too low in the tree. These
        are most often rules like LT02 and LT05 that make whitespace changes
        without a "deep" understanding of the parse structure. This function
        attempts to correct those issues automatically. It may not be perfect,
        but it should be an improvement over the old behaviour, where rules like
        LT02 often corrupted the parse tree, placing spaces in weird places that
        caused issues with other rules. For more context, see issue #1304.
        """
        if not cls._adjust_anchors:
            return

        for fix in lint_result.fixes:
            if fix.anchor:
                fix.anchor = cls._choose_anchor_segment(
                    # If no parent stack, that means the segment itself is the root
                    (
                        context.parent_stack[0]
                        if context.parent_stack
                        else context.segment
                    ),
                    fix.edit_type,
                    fix.anchor,
                )

    @staticmethod
    def _choose_anchor_segment(
        root_segment: BaseSegment,
        edit_type: str,
        segment: BaseSegment,
        filter_meta: bool = False,
    ) -> BaseSegment:
        """Choose the anchor point for a lint fix, i.e. where to apply the fix.

        From a grammar perspective, segments near the leaf of the tree are
        generally less likely to allow general edits such as whitespace
        insertion.

        This function avoids such issues by taking a proposed anchor point
        (assumed to be near the leaf of the tree) and walking "up" the parse
        tree as long as the ancestor segments have the same start or end point
        (depending on the edit type) as "segment". This newly chosen anchor
        is more likely to be a valid anchor point for the fix.
        """
        if edit_type not in ("create_before", "create_after"):
            return segment

        anchor: BaseSegment = segment
        child: BaseSegment = segment
        path: Optional[list[BaseSegment]] = (
            [ps.segment for ps in root_segment.path_to(segment)]
            if root_segment
            else None
        )
        # If no path found, return the original segment. This shouldn't
        # happen in practice given a correctly-parented tree (the anchor
        # is always a descendant of root_segment), but is kept as a
        # defensive fallback rather than raising.
        if not path:  # pragma: no cover
            linter_logger.debug(
                "No path found from %s to %s. Segment may be in unparsable section. "
                "Returning original segment as anchor.",
                root_segment,
                segment,
            )
            return segment
        for seg in path[::-1]:
            # If the segment allows non code ends, then no problem.
            # We're done. This is usually the outer file segment.
            if seg.can_start_end_non_code:
                linter_logger.debug(
                    "Stopping hoist at %s, as allows non code ends.", seg
                )
                break
            # Which lists of children to check against.
            children_lists: list[list[BaseSegment]] = []
            if filter_meta:
                # Optionally check against filtered (non-meta only) children.
                children_lists.append(
                    [child for child in seg.segments if not child.is_meta]
                )
            # Always check against the full set of children.
            children_lists.append(list(seg.segments))
            children: list[BaseSegment]
            for children in children_lists:
                if edit_type == "create_before" and children[0] is child:
                    linter_logger.debug(
                        "Hoisting anchor from before %s to %s", anchor, seg
                    )
                    anchor = seg
                    assert anchor.raw.startswith(segment.raw)
                    child = seg
                    break
                elif edit_type == "create_after" and children[-1] is child:
                    linter_logger.debug(
                        "Hoisting anchor from after %s to %s", anchor, seg
                    )
                    anchor = seg
                    assert anchor.raw.endswith(segment.raw)
                    child = seg
                    break
        return anchor


@dataclass(frozen=True)
class RuleManifest:
    """Element in the rule register."""

    code: str
    name: str
    description: str
    groups: tuple[str, ...]
    aliases: tuple[str, ...]
    rule_class: type[BaseRule]


@dataclass
class RulePack:
    """A bundle of rules to be applied.

    This contains a set of rules, post filtering but also contains the mapping
    required to interpret any noqa messages found in files.

    The reason for this object is that rules are filtered and instantiated
    into this pack in the main process when running in multi-processing mode so
    that user defined rules can be used without reference issues.

    Attributes:
        rules (:obj:`list` of :obj:`BaseRule`): A filtered list of instantiated
            rules to be applied to a given file.
        reference_map (:obj:`dict`): A mapping of rule references to the codes
            they refer to, e.g. `{"my_ref": {"LT01", "LT02"}}`. The references
            (i.e. the keys) may be codes, groups, aliases or names. The values
            of the mapping are sets of rule codes *only*. This object acts as
            a lookup to be able to translate selectors (which may contain
            diverse references) into a consolidated list of rule codes. This
            mapping contains the full set of rules, rather than just the filtered
            set present in the `rules` attribute.
    """

    rules: list[BaseRule]
    reference_map: dict[str, set[str]]

    def codes(self) -> Iterator[str]:
        """Returns an iterator through the codes contained in the pack."""
        return (r.code for r in self.rules)


class RuleSet:
    """Class to define a ruleset.

    A rule set is instantiated on module load, but the references
    to each of its classes are instantiated at runtime. This means
    that configuration values can be passed to those rules live
    and be responsive to any changes in configuration from the
    path that the file is in.

    Rules should be fetched using the :meth:`get_rulelist` command which
    also handles any filtering (i.e. allowlisting and denylisting).

    New rules should be added to the instance of this class using the
    :meth:`register` decorator. That decorator registers the class, but also
    performs basic type and name-convention checks.

    The code for the rule will be parsed from the name, the description
    from the docstring. The eval function is assumed that it will be
    overridden by the subclass, and the parent class raises an error on
    this function if not overridden.

    """

    def __init__(self, name: str, config_info: dict[str, ConfigInfo]) -> None:
        self.name = name
        self.config_info = config_info
        self._register: dict[str, RuleManifest] = {}

    def _validate_config_options(
        self, config: "FluffConfig", rule_ref: Optional[str] = None
    ) -> None:
        """Ensure that all config options are valid.

        Config options can also be checked for a specific rule e.g CP01.
        """
        rule_config = config.get_section("rules")
        for config_name, info_dict in self.config_info.items():
            config_option = (
                rule_config.get(config_name)
                if not rule_ref
                else rule_config.get(rule_ref).get(config_name)
            )
            valid_options = info_dict.get("validation")
            if (
                valid_options
                and config_option not in valid_options
                and config_option is not None
            ):
                raise SQLFluffUserError(
                    (
                        "Invalid option '{}' for {} configuration. Must be one of {}"
                    ).format(
                        config_option,
                        config_name,
                        valid_options,
                    )
                )

    def register(
        self, cls: type[BaseRule], plugin: Optional["PluginSpec"] = None
    ) -> type[BaseRule]:
        """Decorate a class with this to add it to the ruleset.

        .. code-block:: python

           @myruleset.register
           class Rule_LT01(BaseRule):
               "Description of rule."

               def eval(self, **kwargs):
                   return LintResult()

        We expect that rules are defined as classes with the name `Rule_XXXX`
        where `XXXX` is of the form `LNNN`, where L is a letter (literally L for
        *linting* by default) and N is a three digit number.

        If this receives classes by any other name, then it will raise a
        :exc:`ValueError`.

        """
        code = cls.code

        # Check for code collisions.
        if code in self._register:  # pragma: no cover
            raise ValueError(
                "Rule {!r} has already been registered on RuleSet {!r}!".format(
                    code, self.name
                )
            )

        assert "all" in cls.groups, "Rule {!r} must belong to the 'all' group".format(
            code
        )

        self._register[code] = RuleManifest(
            code=code,
            name=cls.name,
            description=cls.description,
            groups=cls.groups,
            aliases=cls.aliases,
            rule_class=cls,
        )

        # Make sure we actually return the original class
        return cls

    def _expand_rule_refs(
        self, glob_list: list[str], reference_map: dict[str, set[str]]
    ) -> set[str]:
        """Expand a list of rule references into a list of rule codes.

        Returns:
            :obj:`set` of :obj:`str` rule codes.

        """
        expanded_rule_set: set[str] = set()
        for r in glob_list:
            # Is it a direct reference?
            if r in reference_map:
                expanded_rule_set.update(reference_map[r])
            # Otherwise treat as a glob expression on all references.
            # NOTE: We expand _all_ references (i.e. groups, aliases, names
            # AND codes) so that we preserve the most backward compatibility
            # with existing references to legacy codes in config files.
            else:
                matched_refs = fnmatch.filter(reference_map.keys(), r)
                for matched in matched_refs:
                    expanded_rule_set.update(reference_map[matched])
        return expanded_rule_set

    def rule_reference_map(self) -> dict[str, set[str]]:
        """Generate a rule reference map for looking up rules.

        Generate the master reference map. The priority order is:
        codes > names > groups > aliases
        (i.e. if there's a collision between a name and an alias - we assume
        the alias is wrong)
        """
        valid_codes: set[str] = set(self._register.keys())
        reference_map: dict[str, set[str]] = {code: {code} for code in valid_codes}

        # Generate name map.
        name_map: dict[str, set[str]] = {
            manifest.name: {manifest.code}
            for manifest in self._register.values()
            if manifest.name
        }
        # Check collisions.
        name_collisions = set(name_map.keys()) & valid_codes
        if name_collisions:  # pragma: no cover
            # NOTE: This clause is untested, because it's quite hard to actually
            # have a valid name which replicates a valid code. The name validation
            # will probably catch it first.
            rules_logger.warning(
                "The following defined rule names were found which collide "
                "with codes. Those names will not be available for selection: %s",
                name_collisions,
            )
        # Incorporate (with existing references taking precedence).
        reference_map = {**name_map, **reference_map}

        # Generate the group map.
        group_map: DefaultDict[str, set[str]] = defaultdict(set)
        for manifest in self._register.values():
            for group in manifest.groups:
                if group in reference_map:
                    rules_logger.warning(
                        "Rule %s defines group %r which is already defined as a "
                        "name or code of %s. This group will not be available "
                        "for use as a result of this collision.",
                        manifest.code,
                        group,
                        reference_map[group],
                    )
                else:
                    group_map[group].add(manifest.code)
        # Incorporate after all checks are done.
        reference_map = {**group_map, **reference_map}

        # Generate the alias map.
        alias_map: DefaultDict[str, set[str]] = defaultdict(set)
        for manifest in self._register.values():
            for alias in manifest.aliases:
                if alias in reference_map:
                    rules_logger.warning(
                        "Rule %s defines alias %r which is already defined as a "
                        "name, code or group of %s. This alias will "
                        "not be available for use as a result of this collision.",
                        manifest.code,
                        alias,
                        reference_map[alias],
                    )
                else:
                    alias_map[alias].add(manifest.code)
        # Incorporate after all checks are done.
        return {**alias_map, **reference_map}

    def get_rulepack(self, config: "FluffConfig") -> RulePack:
        """Use the config to return the appropriate rules.

        We use the config both for allowlisting and denylisting, but also
        for configuring the rules given the given config.
        """
        # Validate all generic rule configs
        self._validate_config_options(config)

        # Fetch config section:
        rules_config = config.get_section("rules")

        # Generate the master reference map. The priority order is:
        # codes > names > groups > aliases
        # (i.e. if there's a collision between a name and an
        # alias - we assume the alias is wrong.)
        valid_codes: set[str] = set(self._register.keys())
        reference_map = self.rule_reference_map()
        valid_config_lookups = {
            manifest.rule_class.get_config_ref() for manifest in self._register.values()
        }

        # Validate config doesn't try to specify values for unknown rules.
        # NOTE: We _warn_ here rather than error.
        for unexpected_ref in [
            # Filtering to dicts gives us the sections.
            k
            for k, v in rules_config.items()
            if isinstance(v, dict)
            # Only keeping ones we don't expect
            if k not in valid_config_lookups
        ]:
            rules_logger.warning(
                "Rule configuration contain a section for unexpected "
                f"rule {unexpected_ref!r}. These values will be ignored."
            )
            # For convenience (and migration), if we do find a potential match
            # for the reference - add that as a warning.
            # NOTE: We don't actually accept config in these cases, even though
            # we could potentially match - because how to resolve _multiple_
            # matching config sections is ambiguous.
            if unexpected_ref in reference_map:
                referenced_codes = reference_map[unexpected_ref]
                if len(referenced_codes) == 1:
                    referenced_code = list(referenced_codes)[0]
                    referenced_name = self._register[referenced_code].name
                    config_ref = self._register[
                        referenced_code
                    ].rule_class.get_config_ref()
                    rules_logger.warning(
                        "The reference was however found as a match for rule "
                        f"{referenced_code} with name {referenced_name!r}. "
                        "SQLFluff assumes configuration for this rule will "
                        f"be specified in 'sqlfluff:rules:{config_ref}'."
                    )
                elif referenced_codes:
                    rules_logger.warning(
                        "The reference was found as a match for multiple rules: "
                        f"{referenced_codes}. Config should be specified by the "
                        "name of the relevant rule e.g. "
                        "'sqlfluff:rules:capitalisation.keywords'."
                    )

        # The lists here are lists of references, which might be codes,
        # names, aliases or groups.
        # We default the allowlist to all the rules if not set (i.e. not specifying
        # any rules, just means "all the rules").
        allowlist = config.get("rule_allowlist") or list(valid_codes)
        denylist = config.get("rule_denylist") or []

        allowlisted_unknown_rule_codes = [
            r
            for r in allowlist
            # Add valid groups to the register when searching for invalid rules _only_
            if not fnmatch.filter(reference_map.keys(), r)
        ]
        if any(allowlisted_unknown_rule_codes):
            rules_logger.warning(
                "Tried to allowlist unknown rule references: {!r}".format(
                    allowlisted_unknown_rule_codes
                )
            )

        denylisted_unknown_rule_codes = [
            r for r in denylist if not fnmatch.filter(reference_map.keys(), r)
        ]
        if any(denylisted_unknown_rule_codes):  # pragma: no cover
            rules_logger.warning(
                "Tried to denylist unknown rules references: {!r}".format(
                    denylisted_unknown_rule_codes
                )
            )

        keylist = sorted(self._register.keys())

        # First we expand the allowlist and denylist globs
        expanded_allowlist = self._expand_rule_refs(allowlist, reference_map)
        expanded_denylist = self._expand_rule_refs(denylist, reference_map)

        # Then we filter the rules
        keylist = [
            r for r in keylist if r in expanded_allowlist and r not in expanded_denylist
        ]

        # Construct the kwargs for each rule and instantiate in turn.
        instantiated_rules = []
        # Keep only config which isn't a section (for specific rule) (i.e. isn't a dict)
        # We'll handle those directly in the specific rule config section below.
        generic_rule_config = {
            k: v for k, v in rules_config.items() if not isinstance(v, dict)
        }
        for code in keylist:
            kwargs = {}
            rule_class = self._register[code].rule_class
            # Fetch the lookup code for the rule.
            rule_config_ref = rule_class.get_config_ref()
            specific_rule_config = config.get_section(("rules", rule_config_ref))
            if generic_rule_config:
                kwargs.update(generic_rule_config)
            if specific_rule_config:
                # Validate specific rule config before adding
                self._validate_config_options(config, rule_config_ref)
                kwargs.update(specific_rule_config)
            kwargs["code"] = code
            # Allow variable substitution in making the description
            kwargs["description"] = self._register[code].description.format(**kwargs)
            # Instantiate when ready
            instantiated_rules.append(rule_class(**kwargs))

        return RulePack(instantiated_rules, reference_map)

    def copy(self) -> "RuleSet":
        """Return a copy of self with a separate register."""
        new_ruleset = copy.copy(self)
        new_ruleset._register = self._register.copy()
        return new_ruleset



"""Implementation of Rule TQ02."""

from typing import List, Optional

from sqlfluff.core.parser import BaseSegment, KeywordSegment, NewlineSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_TQ02(BaseRule):
    """Procedure bodies with multiple statements should be wrapped in BEGIN/END.

    **Anti-pattern**

    Procedure bodies with multiple statements should be wrapped in BEGIN/END
    for clarity and consistency.

    .. code-block:: sql
       :force:

        CREATE PROCEDURE Reporting.MultipleStatements
        AS
        SELECT
            [ID],
            [DataDate],
            [CaseOutput]
        FROM Table1;

        SELECT
            [ID],
            [DataDate],
            [CaseOutput]
        FROM Table2;

    **Best practice**

    Wrap procedure bodies with multiple statements in BEGIN/END blocks.

    .. code-block:: sql
       :force:

        CREATE PROCEDURE Reporting.MultipleStatements
        AS
        BEGIN
            SELECT
                [ID],
                [DataDate],
                [CaseOutput]
            FROM Table1;

            SELECT
                [ID],
                [DataDate],
                [CaseOutput]
            FROM Table2;
        END
    """

    name = "tsql.procedure_begin_end"
    aliases = ()
    groups = ("all", "tsql")
    crawl_behaviour = SegmentSeekerCrawler({"create_procedure_statement"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Procedure bodies with multiple statements should be wrapped in BEGIN/END."""
        # Rule only applies to T-SQL syntax.
        if context.dialect.name != "tsql":
            return None  # pragma: no cover

        # We are only interested in CREATE/ALTER PROCEDURE statements.
        assert context.segment.is_type("create_procedure_statement")

        # Find the procedure_statement (the body after AS)
        procedure_statement = None
        for seg in context.segment.segments:
            if seg.is_type("procedure_statement"):
                procedure_statement = seg
                break

        if not procedure_statement:
            return None  # pragma: no cover

        # Get direct statement children (excluding whitespace, newlines, meta, etc.)
        statements = [
            seg for seg in procedure_statement.segments if seg.is_type("statement")
        ]

        # If there are fewer than 2 statements, no need for BEGIN/END
        if len(statements) < 2:
            return None

        # Check if the first statement is already a begin_end_block
        if statements[0].segments:
            first_child = statements[0].segments[0]
            if first_child.is_type("begin_end_block"):
                # Already wrapped in BEGIN/END
                return None  # pragma: no cover
        else:
            # Defensive: statement with no segments shouldn't happen in valid parsed SQL
            return None  # pragma: no cover

        # We have multiple statements without BEGIN/END - create a fix
        fixes = self._create_begin_end_fixes(procedure_statement, statements)

        return LintResult(
            anchor=procedure_statement,
            description="Procedure body with multiple statements should be wrapped "
            "in BEGIN/END block.",
            fixes=fixes,
        )

    def _create_begin_end_fixes(
        self, procedure_statement: BaseSegment, statements: List[BaseSegment]
    ) -> List[LintFix]:
        """Create fixes to wrap the procedure body in BEGIN/END."""
        # The strategy: insert BEGIN at the start and END at the end
        # We need to find anchor points in the actual segment list

        # Find first statement segment in procedure_statement.segments
        first_statement_idx = None
        for idx, seg in enumerate(procedure_statement.segments):
            if seg in statements:
                first_statement_idx = idx
                break

        if first_statement_idx is None:
            return []  # pragma: no cover

        # Find last statement segment
        last_statement_idx = None
        for idx in range(len(procedure_statement.segments) - 1, -1, -1):
            seg = procedure_statement.segments[idx]
            if seg in statements:
                last_statement_idx = idx
                break

        if last_statement_idx is None:
            return []  # pragma: no cover

        fixes = []

        # Insert BEGIN before first statement
        begin_keyword = KeywordSegment("BEGIN")
        newline = NewlineSegment()

        fixes.append(
            LintFix.create_before(
                procedure_statement.segments[first_statement_idx],
                [begin_keyword, newline],
            )
        )

        # Insert END after last statement and its terminator (if present)
        # In the current branch, statement_terminator is a peer of statement,
        # so we need to check if there's a terminator after the last statement
        insert_after_seg = procedure_statement.segments[last_statement_idx]

        # Check if there's a statement_terminator after the last statement
        for idx in range(last_statement_idx + 1, len(procedure_statement.segments)):
            seg = procedure_statement.segments[idx]
            if seg.is_type("statement_terminator"):
                insert_after_seg = seg
                break
            elif seg.is_code:
                # Hit another code segment, no terminator after last statement
                break

        end_keyword = KeywordSegment("END")

        # Add newline before END
        fixes.append(
            LintFix.create_after(
                insert_after_seg,
                [newline, end_keyword],
            )
        )

        return fixes



"""Implementation of Rule RF01."""

from dataclasses import dataclass, field
from typing import Optional, cast

from sqlfluff.core.dialects.base import Dialect
from sqlfluff.core.dialects.common import (
    AliasInfo,
    ObjectReferenceLevel,
    ObjectReferencePart,
    extract_possible_multipart_references,
    extract_possible_references,
    iter_raw_references,
)
from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.core.rules.reference import object_ref_matches_table
from sqlfluff.dialects.dialect_ansi import ObjectReferenceSegment
from sqlfluff.utils.analysis.query import Query, Selectable

_START_TYPES = [
    "delete_statement",
    "merge_statement",
    "select_statement",
    "update_statement",
]


@dataclass
class RF01Query(Query):
    """Query with custom RF01 info."""

    aliases: list[AliasInfo] = field(default_factory=list)
    standalone_aliases: list[BaseSegment] = field(default_factory=list)
    parent_stack: tuple[BaseSegment, ...] = field(default_factory=tuple)


class Rule_RF01(BaseRule):
    """References cannot reference objects not present in ``FROM`` clause.

    .. note::

       This rule is disabled by default for Athena, BigQuery, Databricks, DuckDB, Hive,
       Redshift, SOQL and SparkSQL due to the support of things like
       structs and lateral views which trigger false positives. It can be
       enabled with the ``force_enable = True`` flag.

    **Anti-pattern**

    In this example, the reference ``vee`` has not been declared.

    .. code-block:: sql

        SELECT
            vee.a
        FROM foo

    **Best practice**

    Remove the reference.

    .. code-block:: sql

        SELECT
            a
        FROM foo

    """

    name = "references.from"
    aliases = ("L026",)
    groups = ("all", "core", "references")
    config_keywords = ["force_enable"]
    # If any of the parents would have also triggered the rule, don't fire
    # because they will more accurately process any internal references.
    crawl_behaviour = SegmentSeekerCrawler(set(_START_TYPES), allow_recurse=False)

    def _eval(self, context: RuleContext) -> list[LintResult]:
        violations: list[LintResult] = []
        dml_target_table: Optional[list[tuple[str, ...]]] = None
        self.logger.debug("Trigger on: %s", context.segment)
        if not context.segment.is_type("select_statement"):
            # Extract first table reference. This will be the target
            # table in a DML statement.
            table_reference = next(
                context.segment.recursive_crawl("table_reference"), None
            )
            if table_reference:
                dml_target_table = self._table_ref_as_tuple(
                    cast(ObjectReferenceSegment, table_reference), context.dialect
                )

        self.logger.debug("DML Reference Table: %s", dml_target_table)
        # Verify table references in any SELECT statements found in or
        # below context.segment in the parser tree.
        query: RF01Query = RF01Query.from_segment(context.segment, context.dialect)
        query.parent_stack = context.parent_stack
        self._analyze_table_references(
            query, dml_target_table, context.dialect, violations
        )
        return violations

    def _alias_info_as_tuples(
        self, alias_info: AliasInfo, dialect: Dialect
    ) -> list[tuple[str, ...]]:
        result: list[tuple[str, ...]] = []
        if alias_info.aliased:
            result.append((alias_info.ref_str,))
        if alias_info.object_reference:
            result += self._table_ref_as_tuple(
                cast(ObjectReferenceSegment, alias_info.object_reference), dialect
            )
        return result

    def _table_ref_as_tuple(
        self,
        table_reference: ObjectReferenceSegment,
        dialect: Dialect,
    ) -> list[tuple[str, ...]]:
        raw_references = list(iter_raw_references(table_reference, dialect.name))
        return [
            tuple(ref.part for ref in raw_references),
            tuple(ref.segments[0].normalize(ref.part) for ref in raw_references),
        ]

    def _analyze_table_references(
        self,
        query: RF01Query,
        dml_target_table: Optional[list[tuple[str, ...]]],
        dialect: Dialect,
        violations: list[LintResult],
    ) -> None:
        # For each query...
        for selectable in query.selectables:
            select_info = selectable.select_info
            self.logger.debug(
                "Selectable: %s",
                selectable,
            )
            if select_info:
                # Record the available tables.
                query.aliases += select_info.table_aliases
                query.standalone_aliases += select_info.standalone_aliases
                self.logger.debug(
                    "Aliases: %s %s",
                    [alias.ref_str for alias in select_info.table_aliases],
                    [standalone.raw for standalone in select_info.standalone_aliases],
                )

                # Try and resolve each reference to a value in query.aliases (or
                # in an ancestor query).
                for r in select_info.reference_buffer:
                    if not self._should_ignore_reference(r, selectable):
                        # This function walks up the query's parent stack if necessary.
                        violation = self._resolve_reference(
                            r, self._get_table_refs(r, dialect), dml_target_table, query
                        )
                        if violation:
                            violations.append(violation)

        # Visit children.
        for child in query.children:
            self._analyze_table_references(
                cast(RF01Query, child), dml_target_table, dialect, violations
            )

    def _should_ignore_reference(
        self, reference: ObjectReferenceSegment, selectable: Selectable
    ) -> bool:
        if self._is_sequence_pseudocolumn(reference, selectable):
            return True
        ref_path = selectable.selectable.path_to(reference)
        # Ignore references occurring in an "INTO" clause:
        # - They are table references, not column references.
        # - They are the target table, similar to an INSERT or UPDATE
        #   statement, thus not expected to match a table in the FROM
        #   clause.
        if ref_path:
            return any(
                ps.segment.is_type("into_table_clause", "containstable_segment")
                for ps in ref_path
            )
        else:
            return False  # pragma: no cover

    def _is_sequence_pseudocolumn(
        self, reference: ObjectReferenceSegment, selectable: Selectable
    ) -> bool:
        """Whether reference is a dialect-specific pseudocolumn access.

        Handles two cases:
        - Oracle/Snowflake: ``sequence.NEXTVAL`` and ``sequence.CURRVAL``
          (or ``db.schema.sequence.NEXTVAL``) are sequence pseudocolumn
          access rather than table/column access.
        - Databricks/SparkSQL: ``_metadata.file_path``,
          ``_metadata.file_name``, etc. are virtual column access for
          file-based data sources.
          See: https://docs.databricks.com/en/ingestion/file-metadata-column.html
        """
        dialect_name = selectable.dialect.name

        reference_parts = list(iter_raw_references(reference, dialect_name))
        if len(reference_parts) < 2:
            return False

        # Oracle/Snowflake: last part is NEXTVAL or CURRVAL.
        if dialect_name in ("oracle", "snowflake"):
            last_part = reference_parts[-1]
            # Quoted identifiers should still be treated as ordinary references.
            return bool(last_part.segments) and (
                last_part.segments[0].is_type("naked_identifier")
                and last_part.part.upper() in {"NEXTVAL", "CURRVAL"}
            )

        # Databricks/SparkSQL: first part is _metadata.
        if dialect_name in ("databricks", "sparksql"):
            first_part = reference_parts[0]
            return bool(first_part.segments) and (
                first_part.segments[0].is_type("naked_identifier")
                and first_part.part.upper() == "_METADATA"
            )

        return False

    def _get_table_refs(
        self, ref: ObjectReferenceSegment, dialect: Dialect
    ) -> list[tuple[ObjectReferencePart, tuple[str, ...]]]:
        """Given ObjectReferenceSegment, determine possible table references."""
        tbl_refs: list[tuple[ObjectReferencePart, tuple[str, ...]]] = []
        # First, handle any schema.table references.
        for sr, tr in extract_possible_multipart_references(
            ref,
            levels=[
                ObjectReferenceLevel.SCHEMA,
                ObjectReferenceLevel.TABLE,
            ],
            dialect_name=dialect.name,
        ):
            tbl_refs.append((tr, (sr.part, tr.part)))
            tbl_refs.append(
                (
                    tr,
                    (
                        sr.segments[0].normalize(sr.part),
                        tr.segments[0].normalize(tr.part),
                    ),
                )
            )
        # Maybe check for simple table references. Two cases:
        # - For most dialects, skip this if it's a schema+table reference -- the
        #   reference was specific, so we shouldn't ignore that by looking
        #   elsewhere.)
        # - Always do this in BigQuery. BigQuery table references are frequently
        #   ambiguous because BigQuery SQL supports structures, making some
        #   multi-level "." references impossible to interpret with certainty.
        #   We may need to genericize this code someday to support other
        #   dialects. If so, this check should probably align somehow with
        #   whether the dialect overrides
        #   ObjectReferenceSegment.extract_possible_references().
        if not tbl_refs or dialect.name in ["bigquery"]:
            for tr in extract_possible_references(
                ref, level=ObjectReferenceLevel.TABLE, dialect_name=dialect.name
            ):
                tbl_refs.append((tr, (tr.part,)))
                tbl_refs.append((tr, (tr.segments[0].normalize(tr.part),)))
        return tbl_refs

    def _resolve_reference(
        self,
        r: ObjectReferenceSegment,
        tbl_refs: list[tuple[ObjectReferencePart, tuple[str, ...]]],
        dml_target_table: Optional[list[tuple[str, ...]]],
        query: RF01Query,
    ) -> Optional[LintResult]:
        # Does this query define the referenced table?
        possible_references = [tbl_ref[1] for tbl_ref in tbl_refs]
        targets: list[tuple[str, ...]] = []
        for alias in query.aliases:
            targets += self._alias_info_as_tuples(alias, query.dialect)
        for standalone_alias in query.standalone_aliases:
            targets.append((standalone_alias.raw,))
            targets.append((standalone_alias.raw_normalized(False),))
        distinct_targets = set(tuple(s.upper() for s in t) for t in targets)

        if self._dialect_supports_dot_access(query.dialect):
            # BigQuery supports having multiple aliases in the FROM statement
            # SparkSQL supports directly accessing values in nested array columns
            if len(distinct_targets) == 1 or query.dialect.name in [
                "bigquery",
                "sparksql",
            ]:
                self.force_enable: bool
                if self.force_enable:
                    # Backwards compatibility.
                    # Nowadays "force_enable" is more of "strict" mode,
                    # for dialects with dot access.
                    pass
                else:
                    return None

        targets += self._get_implicit_targets(query)

        if not object_ref_matches_table(possible_references, targets):
            # No. Check the parent query, if there is one.
            if query.parent:
                return self._resolve_reference(
                    r, tbl_refs, dml_target_table, cast(RF01Query, query.parent)
                )
            # No parent query. If there's a DML statement at the root, check its
            # target table or alias.
            elif not dml_target_table or not object_ref_matches_table(
                possible_references, dml_target_table
            ):
                return LintResult(
                    # Return the first segment rather than the string
                    anchor=tbl_refs[0][0].segments[0],
                    description=f"Reference {r.raw!r} refers to table/view "
                    "not found in the FROM clause or found in ancestor "
                    "statement.",
                )

        return None

    def _get_implicit_targets(self, query: RF01Query) -> list[tuple[str, ...]]:
        if query.dialect.name == "sqlite":
            maybe_create_trigger: Optional[BaseSegment] = next(
                (
                    seg
                    for seg in reversed(query.parent_stack)
                    if seg.is_type("create_trigger")
                ),
                None,
            )
            if not maybe_create_trigger:
                return []
            for seg in maybe_create_trigger.segments:
                if seg.is_type("keyword") and seg.raw_upper == "INSERT":
                    return [("new",)]
                elif seg.is_type("keyword") and seg.raw_upper == "UPDATE":
                    return [("new",), ("old",)]
                elif seg.is_type("keyword") and seg.raw_upper == "DELETE":
                    return [("old",)]
                else:
                    pass  # pragma: no cover

        if query.dialect.name == "postgres":
            for seg in reversed(query.parent_stack):
                if seg.is_type("create_policy_statement") or seg.is_type(
                    "alter_policy_statement"
                ):
                    table_reference = next(seg.recursive_crawl("table_reference"), None)
                    if table_reference:
                        return self._table_ref_as_tuple(
                            cast(ObjectReferenceSegment, table_reference),
                            query.dialect,
                        )

        return []

    def _dialect_supports_dot_access(self, dialect: Dialect) -> bool:
        # Athena:
        # https://docs.aws.amazon.com/athena/latest/ug/filtering-with-dot.html
        # BigQuery:
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/operators#field_access_operator
        # Databricks:
        # https://docs.databricks.com/en/sql/language-manual/functions/dotsign.html
        # DuckDB:
        # https://duckdb.org/docs/sql/data_types/struct#retrieving-from-structs
        # Redshift:
        # https://docs.aws.amazon.com/redshift/latest/dg/query-super.html
        # TODO: all doc links to all referenced dialects
        return dialect.name in (
            "athena",
            "bigquery",
            "databricks",
            "duckdb",
            "hive",
            "redshift",
            "soql",
            "sparksql",
        )



"""The definition of a matchable interface."""

import copy
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.parser.context import ParseContext
    from sqlfluff.core.parser.match_result import MatchResult
    from sqlfluff.core.parser.segments import BaseSegment


T = TypeVar("T", bound="Matchable")


class Matchable(ABC):
    """A base object defining the matching interface."""

    # Matchables are also not meta unless otherwise defined
    is_meta = False

    @abstractmethod
    def is_optional(self) -> bool:
        """Return whether this element is optional."""

    @abstractmethod
    def simple(
        self, parse_context: "ParseContext", crumbs: Optional[tuple[str, ...]] = None
    ) -> Optional[tuple[frozenset[str], frozenset[str]]]:
        """Try to obtain a simple response from the matcher.

        Returns:
            None - if not simple.
            Tuple of two sets of strings if simple. The first is a set of
                uppercase raw strings which would match. The second is a set
                of segment types that would match.

        NOTE: the crumbs kwarg is designed to be used by Ref to
        detect recursion.
        """

    @abstractmethod
    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> "MatchResult":
        """Match against this matcher."""

    def copy(self: T, **kwargs: Any) -> T:  # pragma: no cover
        """Copy this Matchable.

        Matchable objects are usually copied during dialect inheritance.
        One dialect might make a copy (usually with some modifications)
        to a dialect element of a parent dialect which it can then use
        itself. This provides a little more modularity in dialect definition.

        NOTE: This method on the base class is not usually used, as the
        base matchable doesn't have any options for customisation. It is
        more frequently used by grammar objects such as Sequence, which
        provide more options for customisation. Those grammar objects should
        redefine this method accordingly.
        """
        return copy.copy(self)

    @abstractmethod
    def cache_key(self) -> str:
        """A string to use for cache keying.

        This string should be unique at the parsing stage such that
        if there has already been a match against this key for a set
        of segments, that we can reuse that match.
        """



"""Implementation of Rule CP04."""

from typing import Optional

from sqlfluff.core.rules import LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.rules.capitalisation.CP01 import Rule_CP01

try:
    import sqlfluffrs

    _HAS_SQLFLUFFRS = True
except ImportError:  # pragma: no cover
    _HAS_SQLFLUFFRS = False


class Rule_CP04(Rule_CP01):
    """Inconsistent capitalisation of boolean/null literal.

    **Anti-pattern**

    In this example, ``null`` and ``false`` are in lower-case whereas ``TRUE`` is in
    upper-case.

    .. code-block:: sql

        select
            a,
            null,
            TRUE,
            false
        from foo

    **Best practice**

    Ensure all literal ``null``/``true``/``false`` literals are consistently
    upper or lower case

    .. code-block:: sql

        select
            a,
            NULL,
            TRUE,
            FALSE
        from foo

        -- Also good

        select
            a,
            null,
            true,
            false
        from foo

    """

    name = "capitalisation.literals"
    aliases = ("L040",)
    is_fix_compatible = True

    crawl_behaviour = SegmentSeekerCrawler({"null_literal", "boolean_literal"})
    _exclude_types = ()
    _exclude_parent_types = ()
    _description_elem = "Boolean/null literals"

    def _eval_rust(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Rust-native CP04 detection over the arena.

        Runs the whole detection in Rust (`sqlfluffrs.cp04_violations` over
        `root._rs_tree`) and maps each `(leaf_index, fixed_raw)` back to its
        Python segment to emit a standard `LintFix`. Returns ``None`` (Python
        fallback) when the Rust extension/arena is unavailable, or for a regex
        word-ignore. See ``BaseRule._eval_rust``/``Rule_CP01._eval_rust``.

        Deliberately its own override, not shared dispatch through CP01's —
        CP04 shares `capitalisation_policy` with CP01, so a policy-only check
        in an inherited hook would silently run CP01's keyword detection here
        instead of falling back to Python.
        """
        if not _HAS_SQLFLUFFRS:
            return None  # pragma: no cover
        if self.name != "capitalisation.literals":
            # Defensive only: no rule currently subclasses CP04, so this is
            # unreachable today — guards against a future subclass silently
            # inheriting this override the way pre-fix CP04 inherited CP01's.
            return None  # pragma: no cover
        if getattr(self, "ignore_words_regex", None):
            return None
        policy = str(getattr(self, "capitalisation_policy"))
        return self._eval_rust_capitalisation(
            context, sqlfluffrs.cp04_violations, policy
        )



"""Implementation of Rule ST03."""

from sqlfluff.core.rules import BaseRule, EvalResultType, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.analysis.query import Query


def _is_data_modifying_cte(cte_query: Query) -> bool:
    cte_def = cte_query.cte_definition_segment
    if cte_def is not None:
        try:
            next(
                cte_def.recursive_crawl(
                    "insert_statement",
                    "update_statement",
                    "delete_statement",
                    "merge_statement",
                    recurse_into=False,
                    no_recursive_seg_type=["select_statement"],
                )
            )
            return True
        except StopIteration:
            return False

    raise NotImplementedError("CTE definition is required")  # pragma: no cover


class Rule_ST03(BaseRule):
    """Query defines a CTE (common-table expression) but does not use it.

    **Anti-pattern**

    Defining a CTE that is not used by the query is harmless, but it means
    the code is unnecessary and could be removed.

    .. code-block:: sql

        WITH cte1 AS (
          SELECT a
          FROM t
        ),
        cte2 AS (
          SELECT b
          FROM u
        )

        SELECT *
        FROM cte1

    **Best practice**

    Remove unused CTEs.

    .. code-block:: sql

        WITH cte1 AS (
          SELECT a
          FROM t
        )

        SELECT *
        FROM cte1
    """

    name = "structure.unused_cte"
    aliases = ("L045",)
    groups = ("all", "core", "structure")
    crawl_behaviour = SegmentSeekerCrawler({"with_compound_statement"})

    def _eval(self, context: RuleContext) -> EvalResultType:
        result = []
        query: Query = Query.from_root(context.segment, dialect=context.dialect)

        # Some dialects (e.g. Postgres) have data-modifying statements in
        # WITH blocks that are always executed regardless of whether they
        # are referenced by the primary query. Do not flag those as unused.
        # https://github.com/sqlfluff/sqlfluff/issues/7084
        non_data_modifying_ctes = {
            k.upper(): k
            for k, cte in query.ctes.items()
            if not _is_data_modifying_cte(cte)
        }

        # Work through all the references in the file, checking off CTES as the
        # are referenced.
        for reference in context.segment.recursive_crawl("table_reference"):
            non_data_modifying_ctes.pop(reference.raw_normalized(False).upper(), None)

        # For any left un-referenced at the end. Raise an issue about them.
        for name in non_data_modifying_ctes.values():
            cte = query.ctes[name]
            result += [
                LintResult(
                    anchor=cte.cte_name_segment,
                    description=f"Query defines CTE "
                    f'"{cte.cte_name_segment.raw}" '
                    f"but does not use it.",
                )
            ]
        return result



"""Defines the LintedDir class.

This stores the idea of a collection of linted files at a single start path

"""

from collections.abc import Iterable
from typing import Optional, TypedDict, Union

from sqlfluff.core.errors import (
    CheckTuple,
    SerializedObject,
    SQLBaseError,
    SQLLintError,
)
from sqlfluff.core.formatter import FormatterInterface
from sqlfluff.core.linter.linted_file import TMP_PRS_ERROR_TYPES, LintedFile
from sqlfluff.core.parser.segments.base import BaseSegment


class LintingRecord(TypedDict):
    """A class to store the linted file statistics."""

    filepath: str
    violations: list[SerializedObject]
    # Things like file length
    statistics: dict[str, int]
    # Raw timings, in seconds, for both rules and steps
    timings: dict[str, float]


class LintedDir:
    """A class to store the idea of a collection of linted files at a single start path.

    A LintedDir may contain files in subdirectories, but they all share
    a common root.

    Importantly, this class also abstracts away from the given LintedFile
    object and allows us to either _keep_ those objects for later use, or
    extract the results from them and allow the original object to be discarded
    and save memory overhead if not required.
    """

    def __init__(self, path: str, retain_files: bool = True) -> None:
        self.files: list[LintedFile] = []
        self.path: str = path
        self.retain_files: bool = retain_files
        # Records
        self._records: list[LintingRecord] = []
        # Stats
        self._num_files: int = 0
        self._num_clean: int = 0
        self._num_unclean: int = 0
        self._num_violations: int = 0
        self.num_unfiltered_tmp_prs_errors: int = 0
        self._unfiltered_tmp_prs_errors_map: dict[str, int] = {}
        self.num_tmp_prs_errors: int = 0
        self.num_unfixable_lint_errors: int = 0
        # Timing
        self.step_timings: list[dict[str, float]] = []
        self.rule_timings: list[tuple[str, str, float]] = []

    def add(self, file: LintedFile) -> None:
        """Add a file to this path.

        This function _always_ updates the metadata tracking, but may
        or may not persist the `file` object itself depending on the
        `retain_files` argument given on instantiation.
        """
        # Generate serialised violations.
        violation_records = sorted(
            # Keep the warnings
            (v.to_dict() for v in file.get_violations(filter_warning=False)),
            # The tuple allows sorting by line number, then position, then code
            key=lambda v: (v["start_line_no"], v["start_line_pos"], v["code"]),
        )

        record: LintingRecord = {
            "filepath": file.path,
            "violations": violation_records,
            "statistics": {
                "source_chars": (
                    len(file.templated_file.source_str) if file.templated_file else 0
                ),
                "templated_chars": (
                    len(file.templated_file.templated_str) if file.templated_file else 0
                ),
                # These are all the segments in the tree
                "segments": (
                    file.tree.count_segments(raw_only=False) if file.tree else 0
                ),
                # These are just the "leaf" nodes of the tree
                "raw_segments": (
                    file.tree.count_segments(raw_only=True) if file.tree else 0
                ),
            },
            "timings": {},
        }

        if file.timings:
            record["timings"] = {
                # linting, parsing, templating etc...
                **file.timings.step_timings,
                # individual rule timings, by code.
                **file.timings.get_rule_timing_dict(),
            }

        self._records.append(record)

        # Update the stats
        self._num_files += 1
        if file.is_clean():
            self._num_clean += 1
        else:
            self._num_unclean += 1
        self._num_violations += file.num_violations()
        _unfiltered_tmp_prs_errors = file.num_violations(
            types=TMP_PRS_ERROR_TYPES,
            filter_ignore=False,
            filter_warning=False,
        )
        self.num_unfiltered_tmp_prs_errors += _unfiltered_tmp_prs_errors
        self._unfiltered_tmp_prs_errors_map[file.path] = _unfiltered_tmp_prs_errors
        self.num_tmp_prs_errors += file.num_violations(
            types=TMP_PRS_ERROR_TYPES,
        )
        self.num_unfixable_lint_errors += file.num_violations(
            types=SQLLintError,
            fixable=False,
        )

        # Append timings if present
        if file.timings:
            self.step_timings.append(file.timings.step_timings)
            self.rule_timings.extend(file.timings.rule_timings)

        # Finally, if set to persist files, do that.
        if self.retain_files:
            self.files.append(file)

    def check_tuples(
        self, raise_on_non_linting_violations: bool = True
    ) -> list[CheckTuple]:
        """Compress all the tuples into one list.

        NB: This is a little crude, as you can't tell which
        file the violations are from. Good for testing though.
        For more control use `check_tuples_by_path`.
        """
        return [
            check_tuple
            for file in self.files
            for check_tuple in file.check_tuples(
                raise_on_non_linting_violations=raise_on_non_linting_violations
            )
        ]

    def check_tuples_by_path(
        self, raise_on_non_linting_violations: bool = True
    ) -> dict[str, list[CheckTuple]]:
        """Fetch all check_tuples from all contained `LintedDir` objects.

        Returns:
            A dict, with lists of tuples grouped by path.
        """
        assert self.retain_files, (
            "cannot `check_tuples_by_path()` without `retain_files`"
        )
        return {
            file.path: file.check_tuples(
                raise_on_non_linting_violations=raise_on_non_linting_violations
            )
            for file in self.files
        }

    def num_violations(
        self,
        types: type[SQLBaseError] | Iterable[type[SQLBaseError]] | None = None,
        fixable: bool | None = None,
    ) -> int:
        """Count the number of violations in the path."""
        return sum(
            file.num_violations(types=types, fixable=fixable) for file in self.files
        )

    def get_violations(
        self, rules: str | tuple[str, ...] | None = None
    ) -> list[SQLBaseError]:
        """Return a list of violations in the path."""
        return [v for file in self.files for v in file.get_violations(rules=rules)]

    def as_records(self) -> list[LintingRecord]:
        """Return the result as a list of dictionaries.

        Each record contains a key specifying the filepath, and a list of violations.
        This method is useful for serialization as all objects will be builtin python
        types (ints, strs).
        """
        return self._records

    def stats(self) -> dict[str, int]:
        """Return a dict containing linting stats about this path."""
        return {
            "files": self._num_files,
            "clean": self._num_clean,
            "unclean": self._num_unclean,
            "violations": self._num_violations,
        }

    def persist_changes(
        self,
        formatter: Optional[FormatterInterface] = None,
        fixed_file_suffix: str = "",
    ) -> dict[str, Union[bool, str]]:
        """Persist changes to files in the given path.

        This also logs the output as we go using the formatter if present.
        """
        assert self.retain_files, "cannot `persist_changes()` without `retain_files`"
        # Run all the fixes for all the files and return a dict
        buffer: dict[str, Union[bool, str]] = {}
        for file in self.files:
            buffer[file.path] = file.persist_tree(
                suffix=fixed_file_suffix, formatter=formatter
            )
        return buffer

    def discard_fixes_for_lint_errors_in_files_with_tmp_or_prs_errors(self) -> None:
        """Discard lint fixes for files with templating or parse errors."""
        if self.num_unfiltered_tmp_prs_errors:
            # Filter serialised versions if present.
            for record in self._records:
                if self._unfiltered_tmp_prs_errors_map[record["filepath"]]:
                    for v_dict in record["violations"]:
                        if v_dict.get("fixes", []):
                            # We're changing a violating with fixes, to one without,
                            # so we need to increment the cache value.
                            self.num_unfixable_lint_errors += 1
                            v_dict["fixes"] = []
            # Filter the full versions if present.
            for linted_file in self.files:
                if self._unfiltered_tmp_prs_errors_map[linted_file.path]:
                    for violation in linted_file.violations:
                        if isinstance(violation, SQLLintError):
                            violation.fixes = []

    @property
    def tree(self) -> Optional[BaseSegment]:
        """A convenience method for when there is only one file and we want the tree."""
        assert self.retain_files, ".tree() cannot be called if `retain_files` is False."
        assert len(self.files) == 1, (
            ".tree() cannot be called when a LintedDir contains more than one file."
        )
        assert self.files, (
            "LintedDir has no parsed files. There is probably a parsing error."
        )
        return self.files[0].tree



"""The Trino dialect.

See https://trino.io/docs/current/language.html
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    AnySetOf,
    Anything,
    BaseSegment,
    Bracketed,
    CodeSegment,
    Dedent,
    Delimited,
    IdentifierSegment,
    Indent,
    LiteralSegment,
    Matchable,
    Nothing,
    OneOf,
    OptionallyBracketed,
    Ref,
    RegexLexer,
    Sequence,
    StringLexer,
    StringParser,
    SymbolSegment,
    TypedParser,
)
from sqlfluff.core.parser.grammar.lookbehind import is_distinct_from_lookbehind
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects.dialect_trino_keywords import (
    trino_reserved_keywords,
    trino_unreserved_keywords,
)

ansi_dialect = load_raw_dialect("ansi")
trino_dialect = ansi_dialect.copy_as(
    "trino",
    formatted_name="Trino",
    docstring="""**Default Casing**: ``lowercase``, although the case
of a reference is used in the result set column label. If a column is defined
using :code:`CREATE TEMPORARY TABLE foo (COL1 int)`, then :code:`SELECT * FROM foo`
returns a column labelled :code:`col1`, however :code:`SELECT COL1 FROM foo`
returns a column labelled :code:`COL1`.

**Quotes**: String Literals: ``''``, Identifiers: ``""``

The dialect for `Trino <https://trino.io/docs/current/>`_.""",
)

# Set the bare functions: https://trino.io/docs/current/functions/datetime.html
trino_dialect.sets("bare_functions").update(
    ["current_date", "current_time", "current_timestamp", "localtime", "localtimestamp"]
)

# Set keywords
trino_dialect.sets("unreserved_keywords").clear()
trino_dialect.update_keywords_set_from_multiline_string(
    "unreserved_keywords", trino_unreserved_keywords
)

trino_dialect.sets("reserved_keywords").clear()
trino_dialect.update_keywords_set_from_multiline_string(
    "reserved_keywords", trino_reserved_keywords
)

trino_dialect.insert_lexer_matchers(
    # Regexp Replace w/ Lambda: https://trino.io/docs/422/functions/regexp.html
    [
        StringLexer("right_arrow", "->", CodeSegment),
        StringLexer("fat_right_arrow", "=>", CodeSegment),
    ],
    before="like_operator",
)

trino_dialect.add(
    RightArrowOperator=StringParser("->", SymbolSegment, type="binary_operator"),
    LambdaArrowSegment=StringParser("->", SymbolSegment, type="lambda_arrow"),
    ExecuteArrowSegment=StringParser("=>", SymbolSegment, type="execute_arrow"),
    StartAngleBracketSegment=StringParser(
        "<", SymbolSegment, type="start_angle_bracket"
    ),
    EndAngleBracketSegment=StringParser(">", SymbolSegment, type="end_angle_bracket"),
    FormatJsonEncodingGrammar=Sequence(
        "FORMAT",
        "JSON",
        Sequence("ENCODING", OneOf("UTF8", "UTF16", "UTF32"), optional=True),
    ),
)

trino_dialect.bracket_sets("angle_bracket_pairs").update(
    [
        ("angle", "StartAngleBracketSegment", "EndAngleBracketSegment", False),
    ]
)

trino_dialect.patch_lexer_matchers(
    [
        RegexLexer(
            "double_quote",
            r'"([^"]|"")*"',
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r'"((?:[^"]|"")*)"', 1),
                "escape_replacements": [(r'""', '"')],
            },
        ),
    ]
)

trino_dialect.replace(
    DateTimeLiteralGrammar=OneOf(
        Sequence(
            OneOf("DATE", "TIME", "TIMESTAMP"),
            TypedParser(
                "single_quote", LiteralSegment, type="date_constructor_literal"
            ),
        ),
        Ref("IntervalExpressionSegment"),
    ),
    LikeGrammar=Sequence("LIKE"),
    # TODO: There are no custom SQL functions in Trino! How to handle this?
    MLTableExpressionSegment=Nothing(),
    FromClauseTerminatorGrammar=OneOf(
        "WHERE",
        "LIMIT",
        "OFFSET",
        Sequence("GROUP", "BY"),
        Sequence("ORDER", "BY"),
        "HAVING",
        "WINDOW",
        Ref("SetOperatorSegment"),
        Ref("WithNoSchemaBindingClauseSegment"),
        Ref("WithDataClauseSegment"),
        "FETCH",
    ),
    OrderByClauseTerminators=OneOf(
        "LIMIT",
        "OFFSET",
        "HAVING",
        # For window functions
        "WINDOW",
        Ref("FrameClauseUnitGrammar"),
        "FETCH",
    ),
    SelectClauseTerminatorGrammar=OneOf(
        Ref(
            "FromKeywordSegment",
            exclude=is_distinct_from_lookbehind,
        ),
        "WHERE",
        Sequence("ORDER", "BY"),
        "LIMIT",
        Ref("SetOperatorSegment"),
        "FETCH",
    ),
    WhereClauseTerminatorGrammar=OneOf(
        "LIMIT",
        "OFFSET",
        Sequence("GROUP", "BY"),
        Sequence("ORDER", "BY"),
        "HAVING",
        "WINDOW",
        "FETCH",
    ),
    HavingClauseTerminatorGrammar=OneOf(
        Sequence("ORDER", "BY"),
        "LIMIT",
        "OFFSET",
        "WINDOW",
        "FETCH",
    ),
    GroupByClauseTerminatorGrammar=OneOf(
        Sequence("ORDER", "BY"),
        "LIMIT",
        "OFFSET",
        "HAVING",
        "WINDOW",
        "FETCH",
    ),
    # NOTE: This block was copy/pasted from dialect_ansi.py with these changes made:
    #  - "PRIOR" keyword removed
    Expression_A_Unary_Operator_Grammar=OneOf(
        Ref(
            "SignedSegmentGrammar",
            exclude=Sequence(Ref("QualifiedNumericLiteralSegment")),
        ),
        Ref("TildeSegment"),
        Ref("NotOperatorGrammar"),
    ),
    PostFunctionGrammar=ansi_dialect.get_grammar("PostFunctionGrammar").copy(
        insert=[
            Ref("WithinGroupClauseSegment"),
            Ref("WithOrdinalityClauseSegment"),
        ],
    ),
    FunctionContentsGrammar=AnyNumberOf(
        Ref("ExpressionSegment"),
        # A Cast-like function
        Sequence(Ref("ExpressionSegment"), "AS", Ref("DatatypeSegment")),
        # Trim function
        Sequence(
            Ref("TrimParametersGrammar"),
            Ref("ExpressionSegment", optional=True, exclude=Ref.keyword("FROM")),
            "FROM",
            Ref("ExpressionSegment"),
        ),
        # An extract-like or substring-like function
        Sequence(
            OneOf(Ref("DatetimeUnitSegment"), Ref("ExpressionSegment")),
            "FROM",
            Ref("ExpressionSegment"),
        ),
        Sequence(
            # Allow an optional distinct keyword here.
            Ref.keyword("DISTINCT", optional=True),
            OneOf(
                # Most functions will be using the delimited route
                # but for COUNT(*) or similar we allow the star segment
                # here.
                Ref("StarSegment"),
                Delimited(Ref("FunctionContentsExpressionGrammar")),
            ),
        ),
        Ref(
            "OrderByClauseSegment"
        ),  # used by string_agg (postgres), group_concat (exasol),listagg (snowflake)..
        # like a function call: POSITION ( 'QL' IN 'SQL')
        Sequence(
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
            "IN",
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
        ),
        # For JSON_QUERY function
        # https://trino.io/docs/current/functions/json.html#json-query
        Sequence(
            Ref("ExpressionSegment"),  # json_input
            Ref("FormatJsonEncodingGrammar", optional=True),
            Ref("CommaSegment"),
            Ref("ExpressionSegment"),  # json_path
            OneOf(
                Sequence("WITHOUT", Ref.keyword("ARRAY", optional=True), "WRAPPER"),
                Sequence(
                    "WITH",
                    OneOf("CONDITIONAL", "UNCONDITIONAL", optional=True),
                    Ref.keyword("ARRAY", optional=True),
                    "WRAPPER",
                ),
                optional=True,
            ),
        ),
        # For JSON_OBJECT function: key : value [, ...]
        # https://trino.io/docs/current/functions/json.html#json-object
        Sequence(
            Delimited(
                Sequence(
                    Ref("ExpressionSegment"),
                    Ref("ColonDelimiterSegment"),
                    Ref("ExpressionSegment"),
                ),
            ),
            Sequence(
                OneOf("NULL", "ABSENT"),
                "ON",
                "NULL",
                optional=True,
            ),
            Sequence(
                "RETURNING",
                Ref("DatatypeSegment"),
                Ref("FormatJsonEncodingGrammar", optional=True),
                optional=True,
            ),
        ),
        # For JSON_ARRAY / JSON_OBJECT functions: { NULL | ABSENT } ON NULL
        # https://trino.io/docs/current/functions/json.html#json-array
        # https://trino.io/docs/current/functions/json.html#json-object
        Sequence(
            OneOf("NULL", "ABSENT"),
            "ON",
            "NULL",
        ),
        # For JSON functions: RETURNING type [FORMAT JSON [ENCODING ...]]
        # https://trino.io/docs/current/functions/json.html#json-array
        Sequence(
            "RETURNING",
            Ref("DatatypeSegment"),
            Ref("FormatJsonEncodingGrammar", optional=True),
        ),
        Ref("IgnoreRespectNullsGrammar"),
        Ref("IndexColumnDefinitionSegment"),
        Ref("EmptyStructLiteralSegment"),
        Ref("ListaggOverflowClauseSegment"),
    ),
    BinaryOperatorGrammar=OneOf(
        Ref("ArithmeticBinaryOperatorGrammar"),
        Ref("StringBinaryOperatorGrammar"),
        Ref("BooleanBinaryOperatorGrammar"),
        Ref("ComparisonOperatorGrammar"),
        # Add arrow operators for functions (e.g. regexp_replace)
        Ref("RightArrowOperator"),
    ),
    AccessorGrammar=AnyNumberOf(
        Ref("ArrayAccessorSegment"),
        # Add in semi structured expressions
        Ref("SemiStructuredAccessorSegment"),
    ),
    # match ANSI's naked identifier casefold, trino is case-insensitive.
    QuotedIdentifierSegment=TypedParser(
        "double_quote", IdentifierSegment, type="quoted_identifier", casefold=str.upper
    ),
    FunctionContentsExpressionGrammar=OneOf(
        Ref("LambdaExpressionSegment"),
        Ref("ExpressionSegment"),
    ),
    TemporaryTransientGrammar=Nothing(),
    PrimaryKeyGrammar=Nothing(),
    ForeignKeyGrammar=Nothing(),
    UniqueKeyGrammar=Nothing(),
    TemporaryGrammar=Nothing(),
)


class PrimitiveTypeSegment(BaseSegment):
    """Primitive type segment.

    See https://trino.io/docs/current/language/types.html
    """

    type = "primitive_type"
    match_grammar = OneOf(
        # Boolean
        "BOOLEAN",
        # Integer
        "TINYINT",
        "SMALLINT",
        "INTEGER",
        "INT",
        "BIGINT",
        # Floating-point
        "REAL",
        "DOUBLE",
        # Fixed-precision
        Sequence(
            "DECIMAL",
            Ref("BracketedArguments", optional=True),
        ),
        # String
        Sequence(
            OneOf("CHAR", "VARCHAR"),
            Ref("BracketedArguments", optional=True),
        ),
        "VARBINARY",
        "JSON",
        # Date and time
        "DATE",
        Ref("TimeWithTZGrammar"),
        # Others
        "IPADDRESS",
        "UUID",
    )


class DatatypeSegment(BaseSegment):
    """Data type segment.

    See https://trino.io/docs/current/language/types.html
    """

    type = "data_type"
    match_grammar = OneOf(
        Ref("PrimitiveTypeSegment"),
        # Structural
        Ref("ArrayTypeSegment"),
        Ref("MapTypeSegment"),
        Ref("RowTypeSegment"),
    )


class RowTypeSegment(ansi.StructTypeSegment):
    """Expression to construct a ROW datatype."""

    match_grammar = Sequence(
        "ROW",
        Ref("RowTypeSchemaSegment", optional=True),
    )


class RowTypeSchemaSegment(BaseSegment):
    """Expression to construct the schema of a ROW datatype."""

    type = "struct_type_schema"
    match_grammar = Bracketed(
        Delimited(  # Comma-separated list of field names/types
            Sequence(
                OneOf(
                    # ParameterNames can look like Datatypes so can't use
                    # Optional=True here and instead do a OneOf in order
                    # with DataType only first, followed by both.
                    Ref("DatatypeSegment"),
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("DatatypeSegment"),
                    ),
                )
            )
        )
    )


class SemiStructuredAccessorSegment(BaseSegment):
    """A semi-structured data accessor segment."""

    type = "semi_structured_expression"
    match_grammar = Sequence(
        Ref("DotSegment"),
        Ref("SingleIdentifierGrammar"),
        Ref("ArrayAccessorSegment", optional=True),
        AnyNumberOf(
            Sequence(
                Ref("DotSegment"),
                Ref("SingleIdentifierGrammar"),
                allow_gaps=True,
            ),
            Ref("ArrayAccessorSegment", optional=True),
            allow_gaps=True,
        ),
        allow_gaps=True,
    )


class OverlapsClauseSegment(BaseSegment):
    """An `OVERLAPS` clause like in `SELECT."""

    type = "overlaps_clause"
    match_grammar: Matchable = Nothing()


class UnorderedSelectStatementSegment(ansi.UnorderedSelectStatementSegment):
    """A `SELECT` statement without any ORDER clauses or later."""

    match_grammar: Matchable = Sequence(
        Ref("SelectClauseSegment"),
        Ref("FromClauseSegment", optional=True),
        Ref("WhereClauseSegment", optional=True),
        Ref("GroupByClauseSegment", optional=True),
        Ref("HavingClauseSegment", optional=True),
        Ref("NamedWindowSegment", optional=True),
    )


class OffsetClauseSegment(ansi.OffsetClauseSegment):
    """An `OFFSET` clause like in `SELECT`.

    https://trino.io/docs/current/sql/select.html
    """

    type = "offset_clause"
    match_grammar: Matchable = Sequence(
        "OFFSET",
        Indent,
        OneOf(
            Ref("NumericLiteralSegment"),
            Ref("ExpressionSegment", exclude=Ref.keyword("ROW")),
        ),
        OneOf("ROW", "ROWS", optional=True),
        Dedent,
    )


class SelectStatementSegment(ansi.SelectStatementSegment):
    """A `SELECT` statement.

    https://trino.io/docs/current/sql/select.html
    """

    match_grammar = UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[
            Ref("OrderByClauseSegment", optional=True),
            Ref("OffsetClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
            Ref("FetchClauseSegment", optional=True),
        ],
        replace_terminators=True,
        terminators=[
            Ref("SetOperatorSegment"),
            Ref("WithNoSchemaBindingClauseSegment"),
            Ref("WithDataClauseSegment"),
        ],
    )


class ValuesClauseSegment(ansi.ValuesClauseSegment):
    """A `VALUES` clause within in `WITH`, `SELECT`, `INSERT`."""

    match_grammar = Sequence(
        "VALUES",
        Delimited(Ref("ExpressionSegment")),
    )


class IntervalExpressionSegment(BaseSegment):
    """An interval representing a span of time.

    https://trino.io/docs/current/language/types.html#interval-year-to-month
    https://trino.io/docs/current/functions/datetime.html#date-and-time-operators
    """

    type = "interval_expression"
    match_grammar = Sequence(
        "INTERVAL",
        Ref("QuotedLiteralSegment"),
        OneOf("YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND"),
    )


class FrameClauseSegment(BaseSegment):
    """A frame clause for window functions.

    https://trino.io/blog/2021/03/10/introducing-new-window-features.html
    """

    type = "frame_clause"

    _frame_extent = OneOf(
        Sequence("CURRENT", "ROW"),
        Sequence(
            OneOf(
                Ref("NumericLiteralSegment"), Ref("DateTimeLiteralGrammar"), "UNBOUNDED"
            ),
            OneOf("PRECEDING", "FOLLOWING"),
        ),
    )

    match_grammar: Matchable = Sequence(
        Ref("FrameClauseUnitGrammar"),
        OneOf(_frame_extent, Sequence("BETWEEN", _frame_extent, "AND", _frame_extent)),
    )


class SetOperatorSegment(BaseSegment):
    """A set operator such as Union, Intersect or Except."""

    type = "set_operator"
    match_grammar: Matchable = OneOf(
        Sequence("UNION", OneOf("DISTINCT", "ALL", optional=True)),
        Sequence(
            OneOf(
                "INTERSECT",
                "EXCEPT",
            ),
            Ref.keyword("ALL", optional=True),
        ),
        exclude=Sequence("EXCEPT", Bracketed(Anything())),
    )


class StatementSegment(ansi.StatementSegment):
    """A generic segment, to any of its child subsegments."""

    type = "statement"
    match_grammar: Matchable = OneOf(
        Ref("AlterTableStatementSegment"),
        Ref("AnalyzeStatementSegment"),
        Ref("CommentOnStatementSegment"),
        Ref("CreateFunctionStatementSegment"),
        Ref("CreateRoleStatementSegment"),
        Ref("CreateSchemaStatementSegment"),
        Ref("CreateTableStatementSegment"),
        Ref("CreateViewStatementSegment"),
        Ref("DeleteStatementSegment"),
        Ref("DescribeStatementSegment"),
        Ref("DropFunctionStatementSegment"),
        Ref("DropRoleStatementSegment"),
        Ref("DropSchemaStatementSegment"),
        Ref("DropTableStatementSegment"),
        Ref("DropViewStatementSegment"),
        Ref("ExplainStatementSegment"),
        Ref("InsertStatementSegment"),
        Ref("MergeStatementSegment"),
        Ref("SelectableGrammar"),
        Ref("SetSchemaStatementSegment"),
        Ref("TransactionStatementSegment"),
        Ref("UpdateStatementSegment"),
        Ref("UseStatementSegment"),
        Ref("SetSessionStatementSegment"),
        terminators=[Ref("DelimiterGrammar")],
    )


class AnalyzeStatementSegment(BaseSegment):
    """An 'ANALYZE' statement.

    As per docs https://trino.io/docs/current/sql/analyze.html
    """

    type = "analyze_statement"
    match_grammar = Sequence(
        "ANALYZE",
        Ref("TableReferenceSegment"),
        Sequence(
            "WITH",
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("ExpressionSegment"),
                    ),
                ),
            ),
            optional=True,
        ),
    )


class WithOrdinalityClauseSegment(BaseSegment):
    """A WITH ORDINALITY AS t(name1, name2) clause for CROSS JOIN UNNEST(...).

    https://trino.io/docs/current/sql/select.html#unnest

    Trino supports an optional WITH ORDINALITY clause on UNNEST, which
    adds a numerical ordinality column to the UNNEST result.
    """

    type = "withordinality_clause"
    match_grammar = Sequence(
        "WITH",
        "ORDINALITY",
    )


class WithinGroupClauseSegment(BaseSegment):
    """An WITHIN GROUP clause for window functions.

    https://trino.io/docs/current/functions/aggregate.html#array_agg

    Trino supports an optional FILTER during aggregation that comes
    immediately after the WITHIN GROUP clause.

    https://trino.io/docs/current/functions/aggregate.html#filtering-during-aggregation
    """

    type = "withingroup_clause"
    match_grammar = Sequence(
        "WITHIN",
        "GROUP",
        Bracketed(Ref("OrderByClauseSegment", optional=False)),
        Ref("FilterClauseGrammar", optional=True),
    )


class ListaggOverflowClauseSegment(BaseSegment):
    """ON OVERFLOW clause of listagg function.

    https://trino.io/docs/current/functions/aggregate.html#array_agg
    """

    type = "listagg_overflow_clause"
    match_grammar = Sequence(
        "ON",
        "OVERFLOW",
        OneOf(
            "ERROR",
            Sequence(
                "TRUNCATE",
                Ref("QuotedLiteralSegment", optional=True),
                OneOf("WITH", "WITHOUT", optional=True),
                Ref.keyword("COUNT", optional=True),
            ),
        ),
    )


class ArrayTypeSegment(ansi.ArrayTypeSegment):
    """Prefix for array literals optionally specifying the type."""

    type = "array_type"
    match_grammar = Sequence(
        "ARRAY",
        Ref("ArrayTypeSchemaSegment", optional=True),
    )


class ArrayTypeSchemaSegment(ansi.ArrayTypeSegment):
    """Data type segment of the array.

    Trino supports ARRAY(DATA_TYPE) and ARRAY<DATA_TYPE>
    """

    type = "array_type_schema"
    match_grammar = OneOf(
        Bracketed(
            Ref("DatatypeSegment"),
            bracket_pairs_set="angle_bracket_pairs",
            bracket_type="angle",
        ),
        Bracketed(
            Ref("DatatypeSegment"),
            bracket_pairs_set="bracket_pairs",
            bracket_type="round",
        ),
    )


class GroupByClauseSegment(BaseSegment):
    """A `GROUP BY` clause like in `SELECT`."""

    type = "groupby_clause"

    match_grammar: Matchable = Sequence(
        "GROUP",
        "BY",
        Indent,
        OneOf(
            "ALL",
            Ref("CubeRollupClauseSegment"),
            # Add GROUPING SETS support
            Ref("GroupingSetsClauseSegment"),
            Sequence(
                Delimited(
                    OneOf(
                        Ref("ColumnReferenceSegment"),
                        # Can `GROUP BY 1`
                        Ref("NumericLiteralSegment"),
                        # Can `GROUP BY coalesce(col, 1)`
                        Ref("ExpressionSegment"),
                    ),
                    terminators=[Ref("GroupByClauseTerminatorGrammar")],
                ),
            ),
        ),
        Dedent,
    )


class CommentOnStatementSegment(BaseSegment):
    """`COMMENT ON` statement.

    https://trino.io/docs/current/sql/comment.html
    """

    type = "comment_clause"

    match_grammar = Sequence(
        "COMMENT",
        "ON",
        Sequence(
            OneOf(
                Sequence(
                    OneOf(
                        "TABLE",
                        # TODO: Create a ViewReferenceSegment
                        "VIEW",
                    ),
                    Ref("TableReferenceSegment"),
                ),
                Sequence(
                    "COLUMN",
                    # TODO: Does this correctly emit a Table Reference?
                    Ref("ColumnReferenceSegment"),
                ),
            ),
            Sequence("IS", OneOf(Ref("QuotedLiteralSegment"), "NULL")),
        ),
    )


class LambdaExpressionSegment(BaseSegment):
    """Lambda function used in a function."""

    type = "lambda_function"
    match_grammar = Sequence(
        OneOf(
            Ref("ParameterNameSegment"),
            Bracketed(Delimited(Ref("ParameterNameSegment"))),
        ),
        Ref("LambdaArrowSegment"),
        Ref("ExpressionSegment"),
    )


class CreateTableStatementSegment(ansi.CreateTableStatementSegment):
    """A `CREATE TABLE` statement."""

    type = "create_table_statement"
    match_grammar: Matchable = Sequence(
        "CREATE",
        Ref("OrReplaceGrammar", optional=True),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        # Columns and comment syntax:
        Bracketed(
            Delimited(
                Ref("ColumnReferenceSegment"),
                Ref("ColumnDefinitionSegment"),
                Sequence(
                    "LIKE",
                    Ref("TableReferenceSegment"),
                    Sequence(
                        OneOf("INCLUDING", "EXCLUDING"),
                        "PROPERTIES",
                        optional=True,
                    ),
                ),
            ),
            optional=True,
        ),
        Ref("CommentClauseSegment", optional=True),
        Sequence(
            "WITH",
            Bracketed(
                Delimited(
                    Ref("ParameterNameSegment"),
                    Ref("EqualsSegment"),
                    Ref("ExpressionSegment"),
                ),
            ),
            optional=True,
        ),
        # Create AS syntax:
        Sequence(
            "AS",
            OptionallyBracketed(Ref("SelectableGrammar")),
            optional=True,
        ),
        # Create like syntax
        Sequence("WITH", Ref.keyword("NO", optional=True), "DATA", optional=True),
    )


class ColumnDefinitionSegment(ansi.ColumnDefinitionSegment):
    """A column definition, e.g. for CREATE TABLE or ALTER TABLE."""

    type = "column_definition"
    match_grammar: Matchable = Sequence(
        Ref("SingleIdentifierGrammar"),  # Column name
        Ref("DatatypeSegment"),  # Column type
        AnySetOf(
            Sequence(
                "NOT",
                "NULL",
            ),
            Ref("CommentClauseSegment"),
            Sequence(
                "WITH",
                Bracketed(
                    Delimited(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("ExpressionSegment"),
                    ),
                ),
            ),
        ),
    )


class TransactionStatementSegment(ansi.TransactionStatementSegment):
    """A `COMMIT`, `ROLLBACK` or `TRANSACTION` statement.

    https://trino.io/docs/current/sql/commit.html
    https://trino.io/docs/current/sql/rollback.html
    https://trino.io/docs/current/sql/start-transaction.html
    """

    type = "transaction_statement"
    match_grammar: Matchable = OneOf(
        Sequence(
            "COMMIT",
            Ref.keyword("WORK", optional=True),
        ),
        Sequence(
            "ROLLBACK",
            Ref.keyword("WORK", optional=True),
        ),
        Sequence(
            "START",
            "TRANSACTION",
            Delimited(
                Sequence(
                    "ISOLATION",
                    "LEVEL",
                    OneOf(
                        Sequence("READ", "UNCOMMITTED"),
                        Sequence("READ", "COMMITTED"),
                        Sequence("REPEATABLE", "READ"),
                        "SERIALIZABLE",
                    ),
                ),
                Sequence(
                    "READ",
                    OneOf("ONLY", "WRITE"),
                ),
                optional=True,
            ),
        ),
    )


class InsertStatementSegment(ansi.InsertStatementSegment):
    """An `INSERT` statement.

    https://trino.io/docs/current/sql/insert.html
    """

    type = "insert_statement"
    match_grammar: Matchable = Sequence(
        "INSERT",
        "INTO",
        Ref("TableReferenceSegment"),
        OneOf(
            Ref("SelectableGrammar"),
            Sequence(
                Ref("BracketedColumnReferenceListGrammar"),
                Ref("SelectableGrammar"),
            ),
            Ref("DefaultValuesGrammar"),
        ),
    )


class SetSessionStatementSegment(BaseSegment):
    """A `SET SESSION` statement.

    https://trino.io/docs/current/sql/set-session.html
    """

    type = "set_session_statement"
    match_grammar: Matchable = Sequence(
        "SET",
        "SESSION",
        Sequence(
            Ref("ParameterNameSegment"),
            Ref("DotSegment"),
            optional=True,
        ),
        Ref("ParameterNameSegment"),
        Ref("EqualsSegment"),
        Ref("ExpressionSegment"),
    )


class AlterTableStatementSegment(ansi.AlterTableStatementSegment):
    """A `ALTER TABLE` statement.

    https://trino.io/docs/current/sql/alter-table.html
    """

    match_grammar = Sequence(
        "ALTER",
        "TABLE",
        Ref("IfExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            Sequence(
                "RENAME",
                "TO",
                Ref("TableReferenceSegment"),
            ),
            Sequence(
                "ADD",
                "COLUMN",
                Ref("IfNotExistsGrammar", optional=True),
                Ref("ColumnDefinitionSegment"),
                OneOf(
                    "FIRST",
                    "LAST",
                    Sequence(
                        "AFTER",
                        Ref("ColumnReferenceSegment"),
                    ),
                    optional=True,
                ),
            ),
            Sequence(
                "DROP",
                "COLUMN",
                Ref("IfExistsGrammar", optional=True),
                Ref("ColumnReferenceSegment"),
            ),
            Sequence(
                "RENAME",
                "COLUMN",
                Ref("IfExistsGrammar", optional=True),
                Ref("ColumnReferenceSegment"),
                "TO",
                Ref("ColumnReferenceSegment"),
            ),
            Sequence(
                "ALTER",
                "COLUMN",
                Ref("ColumnReferenceSegment"),
                OneOf(
                    Sequence(
                        "SET",
                        "DATA",
                        "TYPE",
                        Ref("DatatypeSegment"),
                    ),
                    Sequence(
                        "DROP",
                        "NOT",
                        "NULL",
                    ),
                ),
            ),
            Sequence(
                "SET",
                "AUTHORIZATION",
                OneOf(
                    Ref("RoleReferenceSegment"),
                    Sequence(
                        "USER",
                        Ref("RoleReferenceSegment"),
                    ),
                    Sequence(
                        "ROLE",
                        Ref("RoleReferenceSegment"),
                    ),
                ),
            ),
            Sequence(
                "SET",
                "PROPERTIES",
                Delimited(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("ExpressionSegment"),
                    ),
                ),
            ),
            Sequence(
                "EXECUTE",
                Ref("FunctionNameSegment"),
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ParameterNameSegment"),
                            Ref("ExecuteArrowSegment"),
                            Ref("ExpressionSegment"),
                        ),
                    ),
                    optional=True,
                ),
                Sequence(
                    "WHERE",
                    Ref("ExpressionSegment"),
                    optional=True,
                ),
            ),
        ),
    )


class MapTypeSegment(BaseSegment):
    """Expression to construct a MAP datatype."""

    type = "map_type"
    match_grammar = Sequence(
        "MAP",
        Ref("MapTypeSchemaSegment", optional=True),
    )


class MapTypeSchemaSegment(BaseSegment):
    """Expression to construct the schema of a MAP datatype."""

    type = "map_type_schema"
    match_grammar = OneOf(
        Bracketed(
            Sequence(
                Ref("PrimitiveTypeSegment"),
                Ref("CommaSegment"),
                Ref("DatatypeSegment"),
            ),
            bracket_pairs_set="angle_bracket_pairs",
            bracket_type="angle",
        ),
        Bracketed(
            Sequence(
                Ref("PrimitiveTypeSegment"),
                Ref("CommaSegment"),
                Ref("DatatypeSegment"),
            ),
            bracket_pairs_set="bracket_pairs",
            bracket_type="round",
        ),
    )



"""Implementation of Rule CV07."""

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.utils.functional import Segments, sp


class Rule_CV07(BaseRule):
    """Top-level statements should not be wrapped in brackets.

    **Anti-pattern**

    A top-level statement is wrapped in brackets.

    .. code-block:: sql
       :force:

        (SELECT
            foo
        FROM bar)

        -- This also applies to statements containing a sub-query.

        (SELECT
            foo
        FROM (SELECT * FROM bar))

    **Best practice**

    Don't wrap top-level statements in brackets.

    .. code-block:: sql
       :force:

        SELECT
            foo
        FROM bar

        -- Likewise for statements containing a sub-query.

        SELECT
            foo
        FROM (SELECT * FROM bar)
    """

    name = "convention.statement_brackets"
    aliases = ("L053",)
    groups = ("all", "convention")
    crawl_behaviour = RootOnlyCrawler()
    is_fix_compatible = True

    @staticmethod
    def _iter_statements(file_segment):
        """Designed to be used on files.

        Yields only direct children, or children of batches.
        """
        for seg in file_segment.segments:
            if seg.is_type("batch"):
                for subseg in seg.segments:
                    if subseg.is_type("statement"):
                        yield subseg
            elif seg.is_type("statement"):
                yield seg

    @classmethod
    def _iter_bracketed_statements(cls, file_segment):
        for stmt in cls._iter_statements(file_segment):
            for seg in stmt.segments:
                if seg.is_type("bracketed"):
                    yield stmt, seg

    def _eval(self, context: RuleContext) -> list[LintResult]:
        """Top-level statements should not be wrapped in brackets."""
        # Because of the root_only_crawler, this can control its own
        # crawling behaviour.
        results = []
        for parent, bracketed_segment in self._iter_bracketed_statements(
            context.segment
        ):
            self.logger.debug("Evaluating %s in %s", bracketed_segment, parent)
            # Replace the bracketed segment with it's
            # children, excluding the bracket symbols.
            bracket_set = {"start_bracket", "end_bracket"}

            filtered_children = Segments(
                *[
                    segment
                    for segment in bracketed_segment.segments
                    if segment.get_type() not in bracket_set and not segment.is_meta
                ]
            )

            # Lift leading/trailing whitespace and inline comments to the
            # segment above. This avoids introducing a parse error (ANSI and other
            # dialects generally don't allow this at lower levels of the parse
            # tree).
            to_lift_predicate = sp.or_(sp.is_whitespace(), sp.is_type("inline_comment"))
            leading = filtered_children.select(loop_while=to_lift_predicate)
            self.logger.debug("Leading: %s", leading)
            trailing = (
                filtered_children.reversed()
                .select(loop_while=to_lift_predicate)
                .reversed()
            )
            self.logger.debug("Trailing: %s", trailing)
            # Order-preserving dedupe (leading and trailing can overlap when
            # everything inside the brackets is liftable). A ``set`` here made
            # the DELETE fix order — and therefore the serialised lint record
            # — depend on segment hashes, which are PYTHONHASHSEED-salted,
            # i.e. nondeterministic across runs.
            lift_nodes = list(dict.fromkeys(leading + trailing))
            fixes = []
            if lift_nodes:
                fixes.append(LintFix.create_before(parent, list(leading)))
                fixes.append(LintFix.create_after(parent, list(trailing)))
                fixes.extend([LintFix.delete(segment) for segment in lift_nodes])
                filtered_children = filtered_children[len(leading) : -len(trailing)]

            fixes.append(
                LintFix.replace(
                    bracketed_segment,
                    filtered_children,
                )
            )

            results.append(LintResult(anchor=bracketed_segment, fixes=fixes))
        return results



"""Implementation of Rule LT12."""

from typing import Optional, cast

from sqlfluff.core.helpers.string import get_trailing_whitespace_from_string
from sqlfluff.core.parser import BaseSegment, NewlineSegment
from sqlfluff.core.parser.segments import SourceFix, TemplateSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.utils.functional import FunctionalContext, Segments, sp, tsp


def get_trailing_newlines(segment: BaseSegment) -> list[BaseSegment]:
    """Returns list of trailing newlines in the tree."""
    result = []
    for seg in segment.recursive_crawl_all(reverse=True):
        if seg.is_type("newline"):
            result.append(seg)
        if not seg.is_whitespace and not seg.is_type("dedent", "end_of_file"):
            break
    return result


def get_last_segment(segment: Segments) -> tuple[list[BaseSegment], Segments]:
    """Returns rightmost & lowest descendant and its "parent stack"."""
    parent_stack: list[BaseSegment] = []
    while True:
        children = segment.children()
        if children:
            parent_stack.append(segment[0])
            segment = children.last(predicate=sp.not_(sp.is_type("end_of_file")))
        else:
            return parent_stack, segment


class Rule_LT12(BaseRule):
    """Files must end with a single trailing newline.

    **Anti-pattern**

    The content in file does not end with a single trailing newline. The ``$``
    represents end of file.

    .. code-block:: sql
       :force:

        SELECT
            a
        FROM foo$

        -- Ending on an indented line means there is no newline
        -- at the end of the file, the • represents space.

        SELECT
        ••••a
        FROM
        ••••foo
        ••••$

        -- Ending on a semi-colon means the last line is not a
        -- newline.

        SELECT
            a
        FROM foo
        ;$

        -- Ending with multiple newlines.

        SELECT
            a
        FROM foo

        $

    **Best practice**

    Add trailing newline to the end. The ``$`` character represents end of file.

    .. code-block:: sql
       :force:

        SELECT
            a
        FROM foo
        $

        -- Ensuring the last line is not indented so is just a
        -- newline.

        SELECT
        ••••a
        FROM
        ••••foo
        $

        -- Even when ending on a semi-colon, ensure there is a
        -- newline after.

        SELECT
            a
        FROM foo
        ;
        $

    """

    name = "layout.end_of_file"
    # Between 2.0.0 and 2.0.4 we supported had a kebab-case name for this rule
    # so the old name remains here as an alias to enable backward compatibility.
    aliases = ("L009", "layout.end-of-file")
    groups = ("all", "core", "layout")

    targets_templated = True
    # Use the RootOnlyCrawler to only call _eval() ONCE, with the root segment.
    crawl_behaviour = RootOnlyCrawler()
    lint_phase = "post"
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """Files must end with a single trailing newline.

        We only care about the segment and the siblings which come after it
        for this rule, we discard the others into the kwargs argument.

        """
        # We only care about the final segment of the parse tree.
        parent_stack, segment = get_last_segment(FunctionalContext(context).segment)
        self.logger.debug("Found last segment as: %s", segment)
        if not segment:
            # NOTE: Edge case. If the file is totally empty, we won't find a final
            # segment. In this case return without error.
            return None

        # Check whether the final segment is a placeholder. If it is, we need
        # to proceed more carefully.
        _trailing_segment = segment.get()
        assert _trailing_segment
        if _trailing_segment.is_type("placeholder"):
            self.logger.debug("Trailing placeholder detected.")
            _template_segment = cast(TemplateSegment, _trailing_segment)
            _trailing_whitespace = get_trailing_whitespace_from_string(
                _template_segment.source_str
            )
            # Does it already end with a single whitespace?
            if _trailing_whitespace == "\n":
                return None
            # If this segment already has fixes applied, don't try and do it
            # again.
            if _template_segment.source_fixes:
                self.logger.debug("Fixes already applied. Aborting...")
                return None
            # It's not a single newline. Return a fix to make it one.
            _current_stop = _template_segment.pos_marker.source_slice.stop
            source_fix = SourceFix(
                "\n",
                slice(
                    _current_stop - len(_trailing_whitespace),
                    _current_stop,
                ),
                _template_segment.pos_marker.templated_slice,
            )
            return LintResult(
                anchor=_template_segment,
                fixes=[
                    LintFix.replace(
                        _template_segment,
                        [_template_segment.edit(source_fixes=[source_fix])],
                    )
                ],
            )

        trailing_newlines = Segments(*get_trailing_newlines(context.segment))
        trailing_literal_newlines = trailing_newlines
        self.logger.debug(
            "Untemplated trailing newlines: %s", trailing_literal_newlines
        )
        if context.templated_file:
            trailing_literal_newlines = trailing_newlines.select(
                loop_while=lambda seg: sp.templated_slices(
                    seg, context.templated_file
                ).all(tsp.is_slice_type("literal"))
            )
        self.logger.debug("Templated trailing newlines: %s", trailing_literal_newlines)
        if not trailing_literal_newlines:
            # We make an edit to create this segment after the child of the FileSegment.
            if len(parent_stack) == 1:
                fix_anchor_segment = segment[0]
            else:
                fix_anchor_segment = parent_stack[1]
            self.logger.debug("Anchor on: %s", fix_anchor_segment)

            return LintResult(
                anchor=segment[0],
                fixes=[
                    LintFix.create_after(
                        fix_anchor_segment,
                        [NewlineSegment()],
                    )
                ],
            )
        elif len(trailing_literal_newlines) > 1:
            # Delete extra newlines.
            return LintResult(
                anchor=segment[0],
                fixes=[LintFix.delete(d) for d in trailing_literal_newlines[1:]],
            )
        else:
            # Single newline, no need for fix.
            return None



"""Linter class and helper classes."""

from sqlfluff.core.formatter import FormatterInterface
from sqlfluff.core.linter.common import ParsedString, RenderedFile, RuleTuple
from sqlfluff.core.linter.linted_file import LintedFile
from sqlfluff.core.linter.linter import Linter
from sqlfluff.core.linter.linting_result import LintingResult

__all__ = (
    "FormatterInterface",
    "RuleTuple",
    "ParsedString",
    "LintedFile",
    "LintingResult",
    "Linter",
    "RenderedFile",
)



"""Implementation of Rule LT15."""

from typing import List, Optional

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_LT15(BaseRule):
    """Too many consecutive blank lines.

    **Anti-pattern**

    In this example, the maximum number of empty lines inside a statement is set to 0.

    .. code-block:: sql

        SELECT 'a' AS col
        FROM tab


        WHERE x = 4
        ORDER BY y


        LIMIT 5
        ;

    **Best practice**

    .. code-block:: sql

        SELECT 'a' AS col
        FROM tab
        WHERE x = 4
        ORDER BY y
        LIMIT 5
        ;

    """

    name = "layout.newlines"
    groups = ("all", "layout")
    config_keywords = [
        "maximum_empty_lines_between_statements",
        "maximum_empty_lines_inside_statements",
        "maximum_empty_lines_between_batches",
    ]
    crawl_behaviour = SegmentSeekerCrawler(types={"newline"}, provide_raw_stack=True)
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[List[LintResult]]:
        """There should be a maximum number of empty lines."""
        self.maximum_empty_lines_between_statements: int
        self.maximum_empty_lines_inside_statements: int
        self.maximum_empty_lines_between_batches: int
        context_seg = context.segment

        # Determine the appropriate maximum based on context
        # Check if we're inside a statement first (highest priority)
        if any(seg.is_type("statement") for seg in context.parent_stack):
            # If directly inside a with_compound_statement (between CTEs or between
            # the last CTE and the main query), use between_statements limit to avoid
            # conflicts with LT08 which requires blank lines after CTEs.
            if context.parent_stack and context.parent_stack[-1].is_type(
                "with_compound_statement"
            ):
                maximum_empty_lines = self.maximum_empty_lines_between_statements
            else:
                maximum_empty_lines = self.maximum_empty_lines_inside_statements
        # Check if we're inside a batch but not in a statement
        elif any(seg.is_type("batch") for seg in context.parent_stack):
            # Inside a batch (between statements in a batch)
            maximum_empty_lines = self.maximum_empty_lines_between_statements
        # At file level - check dialect to determine if between batches or statements
        elif context.dialect.name == "tsql":
            # In T-SQL at file level, we're between batches
            maximum_empty_lines = self.maximum_empty_lines_between_batches
        else:
            # Default: between statements
            maximum_empty_lines = self.maximum_empty_lines_between_statements

        if len(context.raw_stack) < maximum_empty_lines:  # pragma: no cover
            return None

        for raw_seg in context.raw_stack[-maximum_empty_lines - 1 :]:
            if raw_seg.is_templated or not raw_seg.is_type("newline"):
                return None

        return [
            LintResult(
                anchor=context_seg,
                fixes=[LintFix.delete(context_seg)],
            )
        ]



"""Implementation of Rule ST07."""

from typing import Optional

from sqlfluff.core.dialects.common import iter_raw_references
from sqlfluff.core.parser import (
    BaseSegment,
    IdentifierSegment,
    KeywordSegment,
    SymbolSegment,
    WhitespaceSegment,
)
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.dialects.dialect_ansi import ColumnReferenceSegment
from sqlfluff.utils.analysis.select import get_select_statement_info
from sqlfluff.utils.functional import FunctionalContext, Segments, sp


class Rule_ST07(BaseRule):
    """Prefer specifying join keys instead of using ``USING``.

    .. note::
       This rule was originally taken from the `dbt Style Guide
       <https://github.com/dbt-labs/corp/blob/ main/dbt_style_guide.md>`_
       which notes that:

        Certain warehouses have inconsistencies in ``USING``
        results (specifically Snowflake).

       In fact `dbt removed it from their style guide in February 2022
       <https://github.com/dbt-labs/corp/pull/58>`_. However, some like the
       rule, so for now we will keep it in SQLFluff, but encourage those that
       do not find value in the rule, to turn it off.

    .. note::

       This rule is disabled for ClickHouse as it supports ``USING`` without
       brackets which this rule does not support.

    **Anti-pattern**

    .. code-block:: sql

        SELECT
            table_a.field_1,
            table_b.field_2
        FROM
            table_a
        INNER JOIN table_b USING (id)

    **Best practice**

    Specify the keys directly

    .. code-block:: sql

        SELECT
            table_a.field_1,
            table_b.field_2
        FROM
            table_a
        INNER JOIN table_b
            ON table_a.id = table_b.id

    """

    name = "structure.using"
    aliases = ("L032",)
    groups: tuple[str, ...] = ("all", "structure")
    crawl_behaviour = SegmentSeekerCrawler({"join_clause"})
    is_fix_compatible = True
    _dialects_disabled_by_default = [
        "clickhouse",
    ]

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        if context.dialect.name in self._dialects_disabled_by_default:
            return LintResult()

        """Look for USING in a join clause."""
        segment = FunctionalContext(context).segment
        parent_stack = FunctionalContext(context).parent_stack
        # We are not concerned with non join clauses
        assert context.segment.is_type("join_clause")

        using_anchor = segment.children(sp.is_keyword("using")).first()
        # If there is no evidence of a USING then we exit
        if len(using_anchor) == 0:
            return None

        anchor = using_anchor.get()
        description = "Found USING statement. Expected only ON statements."
        # All returns from here out will be some form of linting error.
        # we prepare the variable here
        unfixable_result = LintResult(
            anchor=anchor,
            description=description,
        )

        tables_in_join = parent_stack.last().children(
            sp.is_type("join_clause", "from_expression_element")
        )

        # We can only safely fix the first join clause
        if segment.get(0) != tables_in_join.get(1):
            return unfixable_result

        parent_select = parent_stack.last(sp.is_type("select_statement")).get()
        if not parent_select:  # pragma: no cover
            return unfixable_result

        select_info = get_select_statement_info(parent_select, context.dialect)
        table_aliases = [
            ta
            for ta in (select_info.table_aliases if select_info else [])
            if ta.ref_str
        ]
        if len(table_aliases) < 2:
            return unfixable_result

        using_col_segments = _extract_cols_from_using(segment, using_anchor)
        using_cols = [seg.raw for seg in using_col_segments]

        # A USING join de-duplicates the join columns, so any reference whose
        # leading identifier is a USING column -- a bare ``field_1`` in the
        # SELECT clause, or a struct access like ``field_1.nested_field`` --
        # relies on that de-duplication. Rewriting to an ON condition would
        # make such references ambiguous and produce invalid SQL, so flag the
        # USING clause but leave it unfixable. Names are compared using the
        # dialect's identifier normalization (which respects quoting and
        # case-folding) rather than a raw upper-case; this also stays
        # conservative when a USING column name happens to match a table
        # alias (a ``field_1.nested_field`` struct access is then still
        # treated as unfixable rather than assumed to be ``table.column``).
        # https://github.com/sqlfluff/sqlfluff/issues/7230
        using_cols_normalized = {seg.raw_normalized() for seg in using_col_segments}
        if select_info:
            for ref in select_info.reference_buffer:
                parts = list(iter_raw_references(ref, context.dialect.name))
                if not parts:  # pragma: no cover
                    continue
                if parts[0].segments[0].raw_normalized() in using_cols_normalized:
                    return unfixable_result

        to_delete, insert_after_anchor = _extract_deletion_sequence_and_anchor(segment)

        # This join connects the base table of its from_expression with the
        # table joined here -- not necessarily the SELECT's first two tables,
        # which diverge when a comma-join precedes this USING join.
        alias_by_element = {
            ta.from_expression_element.uuid: ta
            for ta in table_aliases
            if ta.from_expression_element is not None
        }
        left_element = tables_in_join.get(0)
        right_element = segment.children(sp.is_type("from_expression_element")).get(0)
        table_a = alias_by_element.get(left_element.uuid) if left_element else None
        table_b = alias_by_element.get(right_element.uuid) if right_element else None
        if table_a is None or table_b is None:  # pragma: no cover
            return unfixable_result
        edit_segments = [
            KeywordSegment(raw="ON"),
            WhitespaceSegment(raw=" "),
        ] + _generate_join_conditions(
            table_a.ref_str,
            table_b.ref_str,
            using_cols,
        )

        assert table_a.segment
        assert table_b.segment
        fixes = [
            LintFix.create_before(
                anchor_segment=insert_after_anchor,
                source=[table_a.segment, table_b.segment],
                edit_segments=edit_segments,
            ),
            *[LintFix.delete(seg) for seg in to_delete],
        ]
        return LintResult(
            anchor=anchor,
            description=description,
            fixes=fixes,
        )


def _extract_cols_from_using(join_clause: Segments, using_segs: Segments) -> Segments:
    # First bracket after the USING keyword, then find ids
    return (
        join_clause.children()
        .select(start_seg=using_segs[0], select_if=sp.is_type("bracketed"))
        .first()
        .children(sp.is_type("identifier"))
    )


def _generate_join_conditions(
    table_a_ref: str, table_b_ref: str, columns: list[str]
) -> list[BaseSegment]:
    edit_segments: list[BaseSegment] = []
    for col in columns:
        edit_segments = edit_segments + [
            _create_col_reference(
                table_a_ref,
                col,
            ),
            WhitespaceSegment(raw=" "),
            SymbolSegment(raw="="),
            WhitespaceSegment(raw=" "),
            _create_col_reference(
                table_b_ref,
                col,
            ),
            WhitespaceSegment(raw=" "),
            KeywordSegment(raw="AND"),
            WhitespaceSegment(raw=" "),
        ]

    # Trim the " " "AND" " " at the end
    return edit_segments[:-3]


SequenceAndAnchorRes = tuple[list[BaseSegment], BaseSegment]


def _extract_deletion_sequence_and_anchor(
    join_clause: Segments,
) -> SequenceAndAnchorRes:
    insert_anchor: Optional[BaseSegment] = None
    to_delete: list[BaseSegment] = []
    for seg in join_clause.children():
        if seg.raw_upper == "USING":
            # Start collecting once we hit USING
            to_delete.append(seg)
            continue

        if len(to_delete) == 0:
            # Skip if we haven't started collecting
            continue

        if to_delete[-1].is_type("bracketed"):
            # terminate when we hit the brackets
            insert_anchor = seg
            break

        to_delete.append(seg)

    assert insert_anchor, "Insert Anchor must be present at this point"
    return to_delete, insert_anchor


def _create_col_reference(table_ref: str, column_name: str) -> ColumnReferenceSegment:
    segments = (
        IdentifierSegment(raw=table_ref, type="naked_identifier"),
        SymbolSegment(raw=".", type="symbol"),
        IdentifierSegment(raw=column_name, type="naked_identifier"),
    )
    return ColumnReferenceSegment(segments=segments, pos_marker=None)



"""Implementation of Rule LT09."""

from collections.abc import Sequence
from typing import NamedTuple, Optional

from sqlfluff.core.parser import BaseSegment, NewlineSegment, WhitespaceSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import FunctionalContext, Segments, sp


class SelectTargetsInfo(NamedTuple):
    """Info about select targets and nearby whitespace."""

    select_idx: int
    first_new_line_idx: int
    first_select_target_idx: int
    first_whitespace_idx: int
    comment_after_select_idx: int
    select_targets: Sequence[BaseSegment]
    from_segment: Optional[BaseSegment]
    pre_from_whitespace: list[BaseSegment]


class Rule_LT09(BaseRule):
    """Select targets should be on a new line unless there is only one select target.

    .. note::
       By default, a wildcard (e.g. ``SELECT *``) is considered a single select target.
       If you want it to be treated as multiple select targets, configure
       ``wildcard_policy = multiple``.

    .. note::
       By default (``single_target_policy = same_line``), a single select target
       is allowed to remain on the same line as the ``SELECT`` keyword (e.g.
       ``SELECT a FROM foo``). If you want *all* select targets, including single
       ones, to be placed on a new line set ``single_target_policy = new_line``.
       This gives consistent formatting regardless of the number of select targets.

    **Anti-pattern**

    Multiple select targets on the same line.

    .. code-block:: sql

        select a, b
        from foo;

        -- Single select target on its own line
        -- (with default ``single_target_policy = same_line``).

        SELECT
            a
        FROM foo;


    **Best practice**

    Multiple select targets each on their own line.

    .. code-block:: sql

        select
            a,
            b
        from foo;

        -- Single select target on the same line as the ``SELECT``
        -- keyword (with default ``single_target_policy = same_line``).

        SELECT a
        FROM foo;

        -- With ``single_target_policy = new_line``, single select
        -- targets must also be on a new line for consistency.

        SELECT
            a
        FROM foo;

        -- When select targets span multiple lines, however they
        -- can still be on a new line.

        SELECT
            SUM(
                1 + SUM(
                    2 + 3
                )
            ) AS col
        FROM test_table;

    """

    name = "layout.select_targets"
    aliases = ("L036",)
    groups = ("all", "layout")
    config_keywords = ["wildcard_policy", "single_target_policy"]
    crawl_behaviour = SegmentSeekerCrawler({"select_clause"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        self.wildcard_policy: str
        self.single_target_policy: str
        assert context.segment.is_type("select_clause")
        select_targets_info = self._get_indexes(context)
        select_clause = FunctionalContext(context).segment
        wildcards = select_clause.children(
            sp.is_type("select_clause_element")
        ).children(sp.is_type("wildcard_expression"))
        has_wildcard = bool(wildcards)
        if len(select_targets_info.select_targets) == 1 and (
            not has_wildcard or self.wildcard_policy == "single"
        ):
            return self._eval_single_select_target_element(
                select_targets_info,
                context,
            )
        elif len(select_targets_info.select_targets):
            return self._eval_multiple_select_target_elements(
                select_targets_info, context.segment
            )
        return None

    @staticmethod
    def _get_indexes(context: RuleContext) -> SelectTargetsInfo:
        children = FunctionalContext(context).segment.children()
        select_targets = children.select(sp.is_type("select_clause_element"))
        first_select_target_idx = children.find(select_targets.get())
        selects = children.select(sp.is_keyword("select"))
        select_idx = children.find(selects.get()) if selects else -1
        newlines = children.select(sp.is_type("newline"))
        first_new_line_idx = children.find(newlines.get()) if newlines else -1
        comment_after_select_idx = -1
        if newlines:
            comment_after_select = children.select(
                sp.is_type("comment"),
                start_seg=selects.get(),
                stop_seg=newlines.get(),
                loop_while=sp.or_(
                    sp.is_type("comment"), sp.is_type("whitespace"), sp.is_meta()
                ),
            )
            if comment_after_select:
                comment_after_select_idx = (
                    children.find(comment_after_select.get())
                    if comment_after_select
                    else -1
                )
        first_whitespace_idx = -1
        if first_new_line_idx != -1:
            # TRICKY: Ignore whitespace prior to the first newline, e.g. if
            # the line with "SELECT" (before any select targets) has trailing
            # whitespace.
            segments_after_first_line = children.select(
                sp.is_type("whitespace"), start_seg=children[first_new_line_idx]
            )
            first_whitespace_idx = children.find(segments_after_first_line.get())

        siblings_post = FunctionalContext(context).siblings_post
        from_segment = siblings_post.first(sp.is_type("from_clause")).first().get()
        pre_from_whitespace = siblings_post.select(
            sp.is_type("whitespace"), stop_seg=from_segment
        )
        return SelectTargetsInfo(
            select_idx,
            first_new_line_idx,
            first_select_target_idx,
            first_whitespace_idx,
            comment_after_select_idx,
            select_targets,
            from_segment,
            list(pre_from_whitespace),
        )

    def _eval_multiple_select_target_elements(
        self, select_targets_info, segment
    ) -> Optional[LintResult]:
        """Multiple select targets. Ensure each is on a separate line."""
        fixes = []
        previous_code = None
        select_clause_raws = Segments(segment).raw_segments
        for i, select_target in enumerate(select_targets_info.select_targets):
            assert select_target.pos_marker
            target_start_line = select_target.pos_marker.working_line_no
            target_initial_code = (
                Segments(select_target).raw_segments.first(sp.is_code()).get()
            )
            assert target_initial_code
            previous_code = (
                select_clause_raws.select(
                    # Get the first code that isn't a comma.
                    select_if=sp.and_(sp.is_code(), sp.not_(sp.raw_is(","))),
                    start_seg=previous_code,
                    stop_seg=target_initial_code,
                )
                .last()
                .get()
            )
            assert previous_code
            assert previous_code.pos_marker
            previous_end_line = previous_code.pos_marker.working_line_no
            self.logger.debug(
                "- Evaluating %s [%s, %s]: Prev ends with: %s",
                select_target,
                previous_end_line,
                target_start_line,
                previous_code,
            )

            # Check whether this target *starts* on the same line that the
            # previous one *ends* on. If they are on the same line, insert a newline.
            if target_start_line == previous_end_line:
                # Find and delete any whitespace before the select target.
                start_seg = select_targets_info.select_idx
                # If any select modifier (e.g. distinct ) is present, start
                # there rather than at the beginning.
                modifier = segment.get_child("select_clause_modifier")
                if modifier:
                    start_seg = segment.segments.index(modifier)

                ws_to_delete = segment.select_children(
                    start_seg=(
                        segment.segments[start_seg]
                        if not i
                        else select_targets_info.select_targets[i - 1]
                    ),
                    select_if=lambda s: s.is_type("whitespace"),
                    loop_while=lambda s: s.is_type("whitespace", "comma") or s.is_meta,
                )
                fixes += [LintFix.delete(ws) for ws in ws_to_delete]
                fixes.append(LintFix.create_before(select_target, [NewlineSegment()]))

            # If we are at the last select target check if the FROM clause
            # is on the same line, and if so move it to its own line.
            if select_targets_info.from_segment:
                if (i + 1 == len(select_targets_info.select_targets)) and (
                    select_target.pos_marker.working_line_no
                    == select_targets_info.from_segment.pos_marker.working_line_no
                ):
                    fixes.extend(
                        [
                            LintFix.delete(ws)
                            for ws in select_targets_info.pre_from_whitespace
                        ]
                    )
                    fixes.append(
                        LintFix.create_before(
                            select_targets_info.from_segment,
                            [NewlineSegment()],
                        )
                    )

        if fixes:
            return LintResult(anchor=segment, fixes=fixes)

        return None

    def _eval_single_select_target_element(
        self, select_targets_info, context: RuleContext
    ):
        select_clause = FunctionalContext(context).segment
        parent_stack = context.parent_stack
        target_idx = select_targets_info.first_select_target_idx
        select_children = select_clause.children()
        target_seg = select_children[target_idx]

        # When single_target_policy is "new_line", a single select target
        # should be treated like multiple targets so the full multi-target
        # layout logic is applied consistently, including any FROM placement
        # fixes when FROM is still on the same line as the target.
        if self.single_target_policy == "new_line":
            return self._eval_multiple_select_target_elements(
                select_targets_info, context.segment
            )

        # Default behavior (single_target_policy = "same_line"):
        # If it's all on one line, then there's no issue.
        if not (
            select_targets_info.select_idx
            < select_targets_info.first_new_line_idx
            < target_idx
        ):
            self.logger.info(
                "Target at index %s is already on a single line.",
                target_idx,
            )
            return None

        # Does the target contain a newline?
        # i.e. even if it's a single element, does it already span more than
        # one line?
        if "newline" in target_seg.descendant_type_set:
            self.logger.info(
                "Target at index %s spans multiple lines so ignoring.",
                target_idx,
            )
            return None

        if select_targets_info.comment_after_select_idx != -1:
            # The SELECT is followed by a comment on the same line. In order
            # to autofix this, we'd need to move the select target between
            # SELECT and the comment and potentially delete the entire line
            # where the select target was (if it is now empty). This is
            # *fairly tricky and complex*, in part because the newline on
            # the select target's line is several levels higher in the
            # parser tree. Hence, we currently don't autofix this. Could be
            # autofixed in the future if/when we have the time.
            return LintResult(anchor=select_clause.get())

        # Prepare the select clause which will be inserted
        insert_buff = [WhitespaceSegment(), target_seg]
        # Delete the first select target from its original location.
        # We'll add it to the right section at the end, once we know
        # what to add.
        initial_deletes = [target_seg]
        # If there's whitespace before it, delete that too.
        if select_children[target_idx - 1].is_type("whitespace"):
            initial_deletes.append(select_children[target_idx - 1])

        # Do we have a modifier?
        modifier: Optional[Segments]
        modifier = select_children.first(sp.is_type("select_clause_modifier"))

        if (
            # Check if the modifier is one we care about
            modifier
            # We only care if it's not already on the first line.
            and select_children.index(modifier.get())
            >= select_targets_info.first_new_line_idx
        ):
            # Prepend it to the insert buffer
            insert_buff = [WhitespaceSegment(), modifier[0]] + insert_buff

            modifier_idx = select_children.index(modifier.get())
            # Delete the whitespace after it (which is two after, thanks to indent)
            if (
                len(select_children) > modifier_idx + 1
                and select_children[modifier_idx + 2].is_whitespace
            ):
                initial_deletes.append(select_children[modifier_idx + 2])

            # Delete the modifier itself
            initial_deletes.append(modifier[0])

            # Set the position marker for removing the preceding
            # whitespace and newline, which we'll use below.
            start_idx = modifier_idx
            start_seg = modifier[0]
        else:
            # Set the position marker for removing the preceding
            # whitespace and newline, which we'll use below.
            start_idx = target_idx
            start_seg = select_children[select_targets_info.first_new_line_idx]

        fixes = [
            # Insert the select_clause in place of the first newline in the
            # Select statement
            LintFix.replace(
                select_children[select_targets_info.first_new_line_idx],
                insert_buff,
            ),
            # Materialise any deletes so far...
            *(LintFix.delete(seg) for seg in initial_deletes),
        ]

        if parent_stack and parent_stack[-1].is_type("select_statement"):
            select_stmt = parent_stack[-1]
            select_clause_idx = select_stmt.segments.index(select_clause.get())
            after_select_clause_idx = select_clause_idx + 1

            if len(select_stmt.segments) > after_select_clause_idx:
                add_newline = True
                to_delete: Sequence[BaseSegment] = [target_seg]
                next_segment = select_stmt.segments[after_select_clause_idx]

                if next_segment.is_type("newline"):
                    # Since we're deleting the newline, we should also delete all
                    # whitespace before it or it will add random whitespace to
                    # following statements. So walk back through the segment
                    # deleting whitespace until you get the previous newline, or
                    # something else.
                    to_delete = select_children.reversed().select(
                        loop_while=sp.is_type("whitespace"),
                        start_seg=select_children[start_idx],
                    )
                    if to_delete:
                        # The select_clause is immediately followed by a
                        # newline. Delete the newline in order to avoid leaving
                        # behind an empty line after fix, *unless* we stopped
                        # due to something other than a newline.
                        delete_last_newline = select_children[
                            start_idx - len(to_delete) - 1
                        ].is_type("newline")

                        # Delete the newline if we decided to.
                        if delete_last_newline:
                            fixes.append(LintFix.delete(next_segment))

                elif next_segment.is_type("whitespace"):
                    # The select_clause has stuff after (most likely a comment)
                    # Delete the whitespace immediately after the select clause
                    # so the other stuff aligns nicely based on where the select
                    # clause started.
                    fixes.append(LintFix.delete(next_segment))

                if to_delete:
                    # Clean up by moving leftover select_clause segments.

                    # Context: Some of the other fixes we make in
                    # _eval_single_select_target_element() leave leftover
                    # child segments that need to be moved to become
                    # *siblings* of the select_clause.
                    move_after_select_clause = select_children.select(
                        start_seg=start_seg,
                        stop_seg=to_delete[-1],
                    )
                    # :TRICKY: Below, we have a couple places where we
                    # filter to guard against deleting the same segment
                    # multiple times -- this is illegal.
                    all_deletes = {
                        fix.anchor for fix in fixes if fix.edit_type == "delete"
                    }
                    for seg in (*to_delete, *move_after_select_clause):
                        if seg not in all_deletes:
                            fixes.append(LintFix.delete(seg))
                            all_deletes.add(seg)

                    if move_after_select_clause or add_newline:
                        fixes.append(
                            LintFix.create_after(
                                select_clause[0],
                                ([NewlineSegment()] if add_newline else [])
                                + list(move_after_select_clause),
                            )
                        )

        return LintResult(
            anchor=select_clause.get(),
            fixes=fixes,
        )
